#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import unittest

import onnx

from onnxruntime.quantization import QuantType
from onnxruntime.quantization.execution_providers.qnn.mixed_precision_overrides_utils import (
    MixedPrecisionTensorQuantOverridesFixer,
)
from onnxruntime.quantization.tensor_quant_overrides import TensorQuantOverridesHelper


class TestMixedPrecisionQuantOverridesFixer(unittest.TestCase):
    def build_test_model_1(self, shape):
        input_0 = onnx.helper.make_tensor_value_info("input_0", onnx.TensorProto.FLOAT, shape)
        input_1 = onnx.helper.make_tensor_value_info("input_1", onnx.TensorProto.FLOAT, shape)
        output_0 = onnx.helper.make_tensor_value_info("output_0", onnx.TensorProto.FLOAT, shape)
        output_1 = onnx.helper.make_tensor_value_info("output_1", onnx.TensorProto.FLOAT, shape)
        output_2 = onnx.helper.make_tensor_value_info("output_2", onnx.TensorProto.FLOAT, shape)

        op1_node = onnx.helper.make_node("Sigmoid", ["input_0"], ["op1_out"], name="op1")
        op2_node = onnx.helper.make_node("Cos", ["input_1"], ["op2_out"], name="op2")
        op3_node = onnx.helper.make_node("Sin", ["op1_out"], ["op3_out"], name="op3")
        op4_node = onnx.helper.make_node("Tanh", ["op2_out"], ["op4_out"], name="op4")
        op5_node = onnx.helper.make_node("Mul", ["op3_out", "op4_out"], ["op5_out"], name="op5")
        op6_node = onnx.helper.make_node("Relu", ["op5_out"], ["output_0"], name="op6")
        op7_node = onnx.helper.make_node("Cos", ["op2_out"], ["output_1"], name="op7")
        op8_node = onnx.helper.make_node("Sigmoid", ["op2_out"], ["output_2"], name="op8")

        graph = onnx.helper.make_graph(
            [
                op1_node,
                op2_node,
                op3_node,
                op4_node,
                op5_node,
                op6_node,
                op7_node,
                op8_node,
            ],
            "mixed_prec_test",
            [input_0, input_1],
            [output_0, output_1, output_2],
        )
        opset_imports = [
            onnx.helper.make_opsetid("", 18),
        ]
        model = onnx.helper.make_model(graph, opset_imports=opset_imports)
        return onnx.shape_inference.infer_shapes(model)

    def test_fixer_1(self):
        shape = (1, 2, 3)
        model = self.build_test_model_1(shape)
        onnx.save_model(model, "model.onnx")

        default_act_qtype = QuantType.QUInt8
        raw_overrides = {"op4_out": [{"quant_type": QuantType.QUInt16}]}
        overrides = TensorQuantOverridesHelper(raw_overrides)
        fixer = MixedPrecisionTensorQuantOverridesFixer.create_from_model(overrides, model, default_act_qtype)
        fixer.apply(default_act_qtype, default_activation_symmetric=False)

        expected = {
            "op2_out": [
                {"quant_type": QuantType.QUInt8, "convert": {"quant_type": QuantType.QUInt16, "recv_nodes": {"op4"}}}
            ],
            "op3_out": [
                {"quant_type": QuantType.QUInt8, "convert": {"quant_type": QuantType.QUInt16, "recv_nodes": {"op5"}}}
            ],
            "op4_out": [{"quant_type": QuantType.QUInt16}],
            "op5_out": [
                {"quant_type": QuantType.QUInt16, "convert": {"quant_type": QuantType.QUInt8, "recv_nodes": {"op6"}}}
            ],
        }
        self.assertDictEqual(overrides.get_dict(), expected)

    def test_fixer_with_symmetric(self):
        shape = (1, 2, 3)
        model = self.build_test_model_1(shape)
        onnx.save_model(model, "model.onnx")

        default_act_qtype = QuantType.QInt8
        raw_overrides = {"op4_out": [{"quant_type": QuantType.QInt16, "symmetric": True}]}
        overrides = TensorQuantOverridesHelper(raw_overrides)
        fixer = MixedPrecisionTensorQuantOverridesFixer.create_from_model(overrides, model, default_act_qtype)
        fixer.apply(default_act_qtype, default_activation_symmetric=False)

        expected = {
            "op2_out": [
                {
                    "quant_type": QuantType.QInt8,
                    "convert": {"quant_type": QuantType.QInt16, "symmetric": True, "recv_nodes": {"op4"}},
                }
            ],
            "op3_out": [
                {
                    "quant_type": QuantType.QInt8,
                    "convert": {"quant_type": QuantType.QInt16, "symmetric": True, "recv_nodes": {"op5"}},
                }
            ],
            "op4_out": [{"quant_type": QuantType.QInt16, "symmetric": True}],
            "op5_out": [
                {
                    "quant_type": QuantType.QInt16,
                    "symmetric": True,
                    "convert": {"quant_type": QuantType.QInt8, "recv_nodes": {"op6"}},
                }
            ],
        }
        self.assertDictEqual(overrides.get_dict(), expected)

    def test_fixer_upgrade_output(self):
        shape = (1, 2, 3)
        model = self.build_test_model_1(shape)
        onnx.save_model(model, "model.onnx")

        default_act_qtype = QuantType.QUInt8
        raw_overrides = {
            "op4_out": [{"quant_type": QuantType.QUInt16}],
            "output_0": [{"quant_type": QuantType.QUInt16}],
        }
        overrides = TensorQuantOverridesHelper(raw_overrides)
        fixer = MixedPrecisionTensorQuantOverridesFixer.create_from_model(overrides, model, default_act_qtype)
        fixer.apply(default_act_qtype, default_activation_symmetric=False)

        expected = {
            "op2_out": [
                {"quant_type": QuantType.QUInt8, "convert": {"quant_type": QuantType.QUInt16, "recv_nodes": {"op4"}}}
            ],
            "op3_out": [
                {"quant_type": QuantType.QUInt8, "convert": {"quant_type": QuantType.QUInt16, "recv_nodes": {"op5"}}}
            ],
            "op4_out": [{"quant_type": QuantType.QUInt16}],
            "op5_out": [{"quant_type": QuantType.QUInt16}],
            "output_0": [{"quant_type": QuantType.QUInt16}],
        }
        self.assertDictEqual(overrides.get_dict(), expected)

    def test_fixer_upgrade_input(self):
        shape = (1, 2, 3)
        model = self.build_test_model_1(shape)
        onnx.save_model(model, "model.onnx")

        default_act_qtype = QuantType.QUInt8
        raw_overrides = {"op4_out": [{"quant_type": QuantType.QUInt16}], "input_0": [{"quant_type": QuantType.QUInt16}]}
        overrides = TensorQuantOverridesHelper(raw_overrides)
        fixer = MixedPrecisionTensorQuantOverridesFixer.create_from_model(overrides, model, default_act_qtype)
        fixer.apply(default_act_qtype, default_activation_symmetric=False)

        expected = {
            "input_0": [{"quant_type": QuantType.QUInt16}],
            "op1_out": [
                {"quant_type": QuantType.QUInt16, "convert": {"quant_type": QuantType.QUInt8, "recv_nodes": {"op3"}}}
            ],
            "op2_out": [
                {"quant_type": QuantType.QUInt8, "convert": {"quant_type": QuantType.QUInt16, "recv_nodes": {"op4"}}}
            ],
            "op3_out": [
                {"quant_type": QuantType.QUInt8, "convert": {"quant_type": QuantType.QUInt16, "recv_nodes": {"op5"}}}
            ],
            "op4_out": [{"quant_type": QuantType.QUInt16}],
            "op5_out": [
                {"quant_type": QuantType.QUInt16, "convert": {"quant_type": QuantType.QUInt8, "recv_nodes": {"op6"}}}
            ],
        }
        self.assertDictEqual(overrides.get_dict(), expected)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
"""
Loads a model and updates the domain of QuantizeLinear and DequantizeLinear nodes to 'com.microsoft'.
Optionally updates zero-points to 16bit data types.

This is used to create models for testing QDQ transformations with the contrib QDQ ops and/or
16-bit ONNX Q/DQ ops.

Usage:
python3 convert_qdq_ops_to_ms_domain.py --input_model <input onnx model> --output_model <output model> --use_16bit_qdq

Models created with this script:
- qdq_with_multi_consumer_dq_nodes.fixed.qdq_contrib.onnx
- qdq_with_multi_consumer_dq_nodes.fixed.qdq16_contrib.onnx
- qdq_with_multi_consumer_dq_nodes.fixed.qdq16.onnx
- fusion/constant_folding_dequantizelinear.qdq_contrib.onnx
- fusion/constant_folding_dequantizelinear.qdq16_contrib.onnx
- fusion/constant_folding_dequantizelinear.qdq16.onnx
- fusion/constant_folding_qdq_node_unit.qdq_contrib.onnx
- fusion/constant_folding_qdq_node_unit.qdq16_contrib.onnx
- fusion/constant_folding_qdq_node_unit.graph_output.qdq_contrib.onnx
- fusion/constant_folding_qdq_node_unit.graph_output.qdq16_contrib.onnx
"""

from __future__ import annotations

import argparse
import os
import struct
import sys

import onnx
from onnx import shape_inference

QDQ_OPS = ("QuantizeLinear", "DequantizeLinear")
QDQ_CONVERT_TYPES = {onnx.TensorProto.UINT8: onnx.TensorProto.UINT16, onnx.TensorProto.INT8: onnx.TensorProto.INT16}
TYPE_TO_STRUCT_LABEL = {
    onnx.TensorProto.UINT8: "B",
    onnx.TensorProto.INT8: "b",
    onnx.TensorProto.UINT16: "H",
    onnx.TensorProto.INT16: "h",
}


def convert_initializer_to_16bits(initializer: onnx.TensorProto, target_type: onnx.TensorProto.DataType):
    byte_order = ">" if sys.byteorder == "big" else "<"
    byte_label = TYPE_TO_STRUCT_LABEL[initializer.data_type]
    short_label = TYPE_TO_STRUCT_LABEL[target_type]

    # Do not support external data
    if initializer.HasField("data_location") and initializer.data_location == onnx.TensorProto.EXTERNAL:
        raise Exception("Do not support initializers with external data")

    # Need to convert raw_data bytes to 16-bit values.
    # NOTE: For tensors that use .int32_data instead of .raw_data, we don't need any special handling
    # other than updating the data type. This is because the upper 24 bits are already cleared to zero.
    if initializer.HasField("raw_data"):
        num_byte_vals = len(initializer.raw_data)

        # Extract 8-bit values as int32s
        int32_vals = struct.unpack(f"{byte_order}{num_byte_vals}{byte_label}", initializer.raw_data)

        # Repack int32 values as 16-bit values
        initializer.raw_data = struct.pack(f"{byte_order}{num_byte_vals}{short_label}", *int32_vals)

    initializer.data_type = target_type


def convert_qdq_op_to_16bit(
    name_to_initializer: dict[str, onnx.TensorProto],
    name_to_values: dict[str, onnx.ValueInfoProto],
    name_to_inputs: dict[str, onnx.ValueInfoProto],
    name_to_outputs: dict[str, onnx.ValueInfoProto],
    node: onnx.NodeProto,
):
    zp_input = node.input[2] if len(node.input) > 2 else None

    if zp_input in name_to_initializer:
        zp_initializer = name_to_initializer[zp_input]

        zp_target_type = QDQ_CONVERT_TYPES.get(zp_initializer.data_type)
        if zp_target_type:
            convert_initializer_to_16bits(zp_initializer, zp_target_type)

        if node.op_type == "DequantizeLinear":
            input0 = node.input[0]

            if input0 in name_to_initializer:
                input_initializer = name_to_initializer[input0]
                input_target_type = QDQ_CONVERT_TYPES.get(input_initializer.data_type)
                if input_target_type:
                    convert_initializer_to_16bits(input_initializer, input_target_type)
            elif input0 in name_to_values:
                input_val = name_to_values[input0]
                input_target_type = QDQ_CONVERT_TYPES.get(input_val.type.tensor_type.elem_type)
                if input_target_type:
                    input_val.type.tensor_type.elem_type = input_target_type
            elif input0 in name_to_inputs:
                input_val = name_to_inputs[input0]
                input_target_type = QDQ_CONVERT_TYPES.get(input_val.type.tensor_type.elem_type)
                if input_target_type:
                    input_val.type.tensor_type.elem_type = input_target_type
        else:
            # QuantizeLinear
            output0 = node.output[0]

            if output0 in name_to_values:
                output_val = name_to_values[output0]
                output_target_type = QDQ_CONVERT_TYPES.get(output_val.type.tensor_type.elem_type)
                if output_target_type:
                    output_val.type.tensor_type.elem_type = output_target_type
            elif output0 in name_to_outputs:
                output_val = name_to_outputs[output0]
                output_target_type = QDQ_CONVERT_TYPES.get(output_val.type.tensor_type.elem_type)
                if output_target_type:
                    output_val.type.tensor_type.elem_type = output_target_type
    else:
        raise Exception("Only support Q/DQ ops with explicit zero-point inputs")


def update_qdq_nodes(graph: onnx.GraphProto, use_16bit_qdq: bool, use_onnx_domain: bool = False):
    name_to_initializer = {initializer.name: initializer for initializer in graph.initializer}
    name_to_values = {value.name: value for value in graph.value_info}
    name_to_inputs = {g_input.name: g_input for g_input in graph.input}
    name_to_outputs = {g_output.name: g_output for g_output in graph.output}

    for node in graph.node:
        # Handle subgraphs:
        for attr in node.attribute:
            if attr.type == onnx.AttributeProto.GRAPH:
                update_qdq_nodes(attr.g, use_16bit_qdq, use_onnx_domain)
            elif attr.type == onnx.AttributeProto.GRAPHS:
                for subgraph in attr.graphs:
                    update_qdq_nodes(subgraph, use_16bit_qdq, use_onnx_domain)

        # Update Q/DQ domains
        if node.op_type in QDQ_OPS:
            node.domain = "com.microsoft" if not use_onnx_domain else ""

            if use_16bit_qdq:
                convert_qdq_op_to_16bit(name_to_initializer, name_to_values, name_to_inputs, name_to_outputs, node)


def main():
    parser = argparse.ArgumentParser(description="Convert Q/DQ ops to com.microsoft domain (or 16-bit)")
    parser.add_argument("--input_model", type=str, required=True, help="Input onnx model path")
    parser.add_argument("--output_model", type=str, required=False, help="Output onnx model path")
    parser.add_argument("--use_onnx_domain", required=False, action="store_true", help="Use ONNX domain for Q/DQ ops")
    parser.add_argument("--use_16bit_qdq", required=False, action="store_true", help="Convert to 16-bit QDQ")

    args = parser.parse_args()

    model = onnx.load(args.input_model)

    has_ms_domain = False
    onnx_opset_version = 1
    for opset in model.opset_import:
        if opset.domain == "com.microsoft":
            has_ms_domain = True
        elif opset.domain == "" or opset.domain == "ai.onnx":
            onnx_opset_version = opset.version

    if not has_ms_domain:
        model.opset_import.extend([onnx.helper.make_opsetid("com.microsoft", 1)])

    if args.use_onnx_domain and args.use_16bit_qdq and onnx_opset_version < 21:
        model = onnx.version_converter.convert_version(model, 21)  # Opset 21 supports 16-bit Q/DQ

    update_qdq_nodes(model.graph, args.use_16bit_qdq, args.use_onnx_domain)
    model = shape_inference.infer_shapes(model)
    onnx.checker.check_model(model, True)

    output_model_path = args.output_model
    if not output_model_path:
        base_model_name = os.path.splitext(args.input_model)[0]
        suffix0 = "" if args.use_onnx_domain else "_contrib"
        suffix = f".qdq16{suffix0}" if args.use_16bit_qdq else f".qdq{suffix0}"
        output_model_path = base_model_name + suffix + ".onnx"

    onnx.save_model(model, output_model_path)
    print(f"[INFO] Saved model: {output_model_path}")


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import numpy as np
import onnx
import onnxscript
from onnx import numpy_helper
from onnxscript import opset17 as op


@onnxscript.script()
def build_model(x: onnxscript.FLOAT):
    y_scale = op.Constant(value_float=1.0)
    y_zero_point = op.Constant(value=numpy_helper.from_array(np.array(1, dtype=np.uint8)))
    return op.Add(x, y_scale), y_scale, y_zero_point


model: onnx.ModelProto = build_model.to_model_proto()
onnx.save(model, "scalar_const_not_share.onnx")



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import onnx
from onnx import TensorProto, helper


def generate_model(model_name):
    nodes = [
        helper.make_node(
            "DisentangledAttention_TRT",
            ["input1", "input2", "input3"],
            ["output"],
            "DisentangledAttention_TRT",
            domain="trt.plugins",
            factor=0.123,
            span=128,
        ),
    ]

    graph = helper.make_graph(
        nodes,
        "trt_plugin_custom_op",
        [  # input
            helper.make_tensor_value_info("input1", TensorProto.FLOAT, [12, 256, 256]),
            helper.make_tensor_value_info("input2", TensorProto.FLOAT, [12, 256, 256]),
            helper.make_tensor_value_info("input3", TensorProto.FLOAT, [12, 256, 256]),
        ],
        [  # output
            helper.make_tensor_value_info("output", TensorProto.FLOAT, [12, 256, 256]),
        ],
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


if __name__ == "__main__":
    generate_model("trt_plugin_custom_op_test.onnx")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from .get_azcopy import get_azcopy  # noqa: F401
from .logger import get_logger
from .platform_helpers import is_linux, is_macOS, is_windows  # noqa: F401
from .run import run  # noqa: F401
from .vcpkg_helpers import (  # noqa: F401
    generate_android_triplets,
    generate_posix_triplets,
    generate_vcpkg_triplets_for_emscripten,
    generate_windows_triplets,
)

try:
    import flatbuffers  # noqa: F401

    from .reduced_build_config_parser import parse_config  # noqa: F401
except ImportError:
    get_logger("tools_python_utils").info("flatbuffers module is not installed. parse_config will not be available")

# see if we can make the pytorch helpers available.
import importlib.util

have_torch = importlib.util.find_spec("torch")
if have_torch:
    from .pytorch_export_helpers import infer_input_info  # noqa: F401



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""This file is used to generate test data for MemoryOptimizer tests in
onnxruntime/test/optimizer/memory_optimizer_test.cc.

Be noticed, after run this script, manually rename recompute_XXXX_execution_model_training.onnx to
recompute_XXXX.onnx
"""

import torch

from onnxruntime.training.ortmodule import DebugOptions, ORTModule


class LinearGeluLinearTest(torch.nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super().__init__()

        self.fc1 = torch.nn.Linear(input_size, hidden_size)
        self.fc2 = torch.nn.Linear(hidden_size, num_classes)

    def forward(self, input1):
        out = self.fc1(input1)
        out = torch.nn.functional.gelu(out)
        out = self.fc2(out)
        return out


DEVICE = "cuda"


def generate_gelu_test_case():
    batch_size, dimension_in, hidden_size, dimension_out = 64, 784, 500, 10
    model = LinearGeluLinearTest(dimension_in, hidden_size, dimension_out).to(DEVICE)
    ort_model = ORTModule(model, DebugOptions(save_onnx=True, onnx_prefix="recompute_gelu"))

    input = torch.randn(batch_size, dimension_in, device=DEVICE)
    # Make sure model runs without any exception
    prediction = ort_model(input)
    assert prediction is not None
    prediction = prediction.sum()
    prediction.backward()


class TileTransposeLinearTest(torch.nn.Module):
    def __init__(self, head):
        super().__init__()
        self._head = head

    # input1 -  float16[24,512,64]
    # repeat -  float16[4]
    # query_layer - float16[24*labels_dim0,512,64]
    def forward(self, input1, query_layer):
        # Tile to [24*labels_dim0,512,64]
        output = input1.repeat(query_layer.size(0) // self._head, 1, 1)

        # Transpose to [24*labels_dim0,64,512]
        output = output.permute(0, 2, 1).contiguous()
        return torch.matmul(query_layer, output)


def generate_tile_test_case():
    batch_size = 16
    head = 24
    seq_length = 512
    model = TileTransposeLinearTest(head).to(DEVICE)
    model = ORTModule(model, DebugOptions(save_onnx=True, onnx_prefix="recompute_tile"))

    input1 = torch.randn(head, seq_length, 64, device=DEVICE).requires_grad_(True)
    query_layer = torch.randn(batch_size * head, seq_length, 64, device=DEVICE).requires_grad_(True)

    # Make sure model runs without any exception
    prediction = model(input1, query_layer)
    assert prediction is not None
    prediction = prediction.sum()

    prediction.backward()


def main():
    """Main entry."""
    generate_gelu_test_case()
    generate_tile_test_case()


if __name__ == "__main__":
    main()



import argparse

import numpy as np
import pandas as pd

ep_map = {"cpu": "CPU", "cuda": "CUDA", "trt": "TRT EP", "native": "Standalone TRT"}


def parse_arguments():
    # create parser
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--prev", required=True, help="previous csv")
    parser.add_argument("-c", "--current", required=True, help="current csv")
    parser.add_argument("-o", "--output_csv", required=True, help="output different csv")
    parser.add_argument(
        "--ep",
        required=False,
        default="trt",
        choices=["cpu", "cuda", "trt", "native"],
        help="ep to capture regressions on",
    )
    parser.add_argument(
        "--tolerance",
        required=False,
        default=0,
        help="allowed tolerance for latency comparison",
    )
    args = parser.parse_args()
    return args


def get_table_condition(table, fp, ep, tol):
    ep = ep_map[ep]
    col1 = ep + " " + fp + " \nmean (ms)_x"
    col2 = ep + " " + fp + " \nmean (ms)_y"
    condition = table[col1] > (table[col2] + tol)
    return condition


def main():
    args = parse_arguments()
    a = pd.read_csv(args.prev)
    b = pd.read_csv(args.current)

    common = a.merge(b, on=["Model"])

    condition_fp32 = get_table_condition(common, "fp32", args.ep, args.tolerance)
    condition_fp16 = get_table_condition(common, "fp16", args.ep, args.tolerance)

    common["greater"] = np.where((condition_fp32 | condition_fp16), True, False)
    greater = common[common["greater"] is True].drop(["greater"], axis=1)

    # arrange columns
    keys = list(greater.keys().sort_values())
    keys.insert(0, keys.pop(keys.index("Model")))
    greater = greater[keys]

    greater.to_csv(args.output_csv)


if __name__ == "__main__":
    main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SparseTensor(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SparseTensor()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSparseTensor(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def SparseTensorBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x4F\x52\x54\x4D", size_prefixed=size_prefixed)

    # SparseTensor
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SparseTensor
    def Values(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from ort_flatbuffers_py.fbs.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SparseTensor
    def Indices(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from ort_flatbuffers_py.fbs.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SparseTensor
    def Dims(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # SparseTensor
    def DimsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # SparseTensor
    def DimsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SparseTensor
    def DimsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def SparseTensorStart(builder):
    builder.StartObject(3)

def Start(builder):
    SparseTensorStart(builder)

def SparseTensorAddValues(builder, values):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(values), 0)

def AddValues(builder, values):
    SparseTensorAddValues(builder, values)

def SparseTensorAddIndices(builder, indices):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(indices), 0)

def AddIndices(builder, indices):
    SparseTensorAddIndices(builder, indices)

def SparseTensorAddDims(builder, dims):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(dims), 0)

def AddDims(builder, dims):
    SparseTensorAddDims(builder, dims)

def SparseTensorStartDimsVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)

def StartDimsVector(builder, numElems: int) -> int:
    return SparseTensorStartDimsVector(builder, numElems)

def SparseTensorEnd(builder):
    return builder.EndObject()

def End(builder):
    return SparseTensorEnd(builder)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
# This script helps evaluation of GPT-2 model.
import logging
import math
import os
import statistics
import timeit

import numpy
import torch
from benchmark_helper import Precision
from gpt2_helper import Gpt2Helper, Gpt2Inputs

logger = logging.getLogger(__name__)


class Gpt2Metric:
    def __init__(self, treatment_name, baseline_name="Torch", top_k=20):
        assert top_k > 1 and top_k <= 100
        self.baseline = baseline_name
        self.treatment = treatment_name
        self.name: str = f"{treatment_name} vs {baseline_name}"
        self.top_k = top_k
        self.top_1_error: int = 0
        self.top_k_error: int = 0
        self.total_samples: int = 0
        self.max_logits_diff: float = 0  # for non-empty past state
        self.max_logits_diff_no_past: float = 0  # for empty past state
        self.batch_top1_error: torch.FloatTensor = None  # top 1 error for current batch
        self.batch_topk_error: torch.FloatTensor = None  # top k error for current batch
        self.seq_len_latency = {}

    def print(self):
        if self.baseline != self.treatment:
            print("---")
            print(f"Metrics for {self.treatment} (baseline={self.baseline}):")
            if self.total_samples > 0:
                top_1_error_rate = 100.0 * self.top_1_error / self.total_samples
                top_k_error_rate = 100.0 * self.top_k_error / self.total_samples
                print(
                    f"Total={self.total_samples} Top1Error={self.top_1_error} ({top_1_error_rate:.2f}%) Top{self.top_k}Error={self.top_k_error} ({top_k_error_rate:.2f}%)"
                )
            print("Max logits diffs:")
            print(f"\twith past  = {self.max_logits_diff:.6f}")
            print(f"\tempty past = {self.max_logits_diff_no_past:.6f}")
        else:
            print(f"Metrics for {self.treatment} (baseline):")

        if self.seq_len_latency:
            print("Past sequence length range and average latency:")
            total = 0
            count = 0
            for key in sorted(self.seq_len_latency.keys()):
                average = statistics.mean(self.seq_len_latency[key]) * 1000.0
                if key == 0:
                    print(f"\t{key}:         \t{average:.2f} ms")
                else:
                    print(f"\t[{2**key}, {2 ** (key + 1) - 1}]:\t{average:.2f} ms")
                total += average * len(self.seq_len_latency[key])
                count += len(self.seq_len_latency[key])
            print(f"Average Latency: {total / count:.2f} ms")

    def diff_logits(self, baseline_logits, treatment_logits, is_empty_past: bool):
        diff = (baseline_logits - treatment_logits).abs().max()
        if is_empty_past:
            self.max_logits_diff_no_past = max(self.max_logits_diff_no_past, diff)
        else:
            self.max_logits_diff = max(self.max_logits_diff, diff)

        return diff

    def start_batch(self, batch_size: int):
        self.total_samples += batch_size
        self.batch_top1_error = torch.zeros((batch_size, 1), dtype=torch.bool)
        self.batch_topk_error = torch.zeros((batch_size, 1), dtype=torch.bool)

    def eval_batch(self, baseline, treatment, past_seq_len, verbose=True):
        self._eval_topk(baseline.top_1_tokens, treatment.top_1_tokens, 1, verbose)
        self._eval_topk(baseline.top_k_tokens, treatment.top_k_tokens, self.top_k, verbose)

        max_diff = self.diff_logits(baseline.logits, treatment.logits, past_seq_len == 0)
        if verbose:
            print(f"Max logits diffs of {self.name}: {max_diff}")

    def _eval_topk(self, baseline_topk, treatment_topk, top_k, verbose=True):
        if not torch.all(torch.eq(baseline_topk, treatment_topk)):
            if top_k == 1:
                if verbose:
                    print(f"Generated tokens not matched for {self.name}")
                self.batch_top1_error |= torch.eq(baseline_topk, treatment_topk).logical_not()
            else:
                if verbose:
                    print(
                        f"Top {top_k} tokens not matched for {self.name}. This will lead to wrong beam search results"
                    )
                self.batch_topk_error |= (
                    torch.eq(baseline_topk, treatment_topk).logical_not().sum(1).unsqueeze(dim=1) > 0
                )

    def end_batch(self):
        self.top_1_error += self.batch_top1_error.sum()
        self.top_k_error += self.batch_topk_error.sum()

    def add_latency(self, past_seq_len, latency):
        key = int(math.log2(past_seq_len)) + 1 if past_seq_len > 0 else 0
        if key not in self.seq_len_latency:
            self.seq_len_latency[key] = []
        self.seq_len_latency[key].append(latency)


class Gpt2Tester:
    def __init__(
        self,
        input_ids,
        position_ids,
        attention_mask,
        num_attention_heads,
        hidden_size,
        num_layer,
        device,
        is_fp16=False,
        top_k=20,
        top_k_required_order=False,
    ):
        self.batch_size = input_ids.shape[0]
        self.input_length = input_ids.shape[1]
        self.n_layer = num_layer

        self.input_ids = input_ids
        self.position_ids = position_ids
        self.attention_mask = attention_mask

        self.has_position_ids = position_ids is not None
        self.has_attention_mask = attention_mask is not None

        # Empty past state for first inference
        self.past = []
        past_shape = [
            2,
            self.batch_size,
            num_attention_heads,
            0,
            hidden_size // num_attention_heads,
        ]
        for _i in range(num_layer):
            empty_past = torch.empty(past_shape).type(torch.float16 if is_fp16 else torch.float32)
            self.past.append(empty_past.to(device))

        self.logits = None
        self.top_1_tokens = None
        self.top_k_tokens = None
        self.top_k = top_k
        self.top_k_required_order = top_k_required_order

    def get_inputs(self) -> Gpt2Inputs:
        return Gpt2Inputs(self.input_ids, self.position_ids, self.attention_mask, self.past)

    def save_test_data(self, session, output, save_test_data_dir, test_case_id):
        from onnx import numpy_helper

        path = os.path.join(save_test_data_dir, "test_data_set_" + str(test_case_id))
        if os.path.exists(path):
            print(f"Directory {path} existed. Skip saving test data")
            return

        os.makedirs(path, exist_ok=True)

        def add_tensor(input_tensors, torch_tensor, name):
            input_tensors.append(numpy_helper.from_array(torch_tensor.clone().cpu().numpy(), name))

        input_tensors = []
        add_tensor(input_tensors, self.input_ids, "input_ids")

        if self.has_position_ids:
            add_tensor(input_tensors, self.position_ids, "position_ids")

        if self.has_attention_mask:
            add_tensor(input_tensors, self.attention_mask, "attention_mask")

        for i in range(self.n_layer):
            add_tensor(input_tensors, self.past[i], "past_" + str(i))

        for i, tensor in enumerate(input_tensors):
            with open(os.path.join(path, f"input_{i}.pb"), "wb") as f:
                f.write(tensor.SerializeToString())

        output_names = [output.name for output in session.get_outputs()]
        for i, _name in enumerate(output_names):
            tensor = numpy_helper.from_array(
                output[i] if isinstance(output[i], numpy.ndarray) else output[i].clone().cpu().numpy()
            )
            with open(os.path.join(path, f"output_{i}.pb"), "wb") as f:
                f.write(tensor.SerializeToString())

        print(f"Test data saved to directory {path}")

    def update(self, output, step, device):
        """
        Update the inputs for next inference.
        """
        self.logits = (
            torch.from_numpy(output[0]) if isinstance(output[0], numpy.ndarray) else output[0].clone().detach().cpu()
        )

        self.top_1_tokens = Gpt2Tester.predict_next_token(self.logits)
        self.top_k_tokens = Gpt2Tester.predict_next_token(self.logits, self.top_k, self.top_k_required_order)

        self.input_ids = self.top_1_tokens.clone().detach().reshape([self.batch_size, 1]).to(device)

        if self.has_position_ids:
            self.position_ids = (
                torch.tensor([self.input_length + step - 1]).unsqueeze(0).repeat(self.batch_size, 1).to(device)
            )

        if self.has_attention_mask:
            self.attention_mask = torch.cat(
                [
                    self.attention_mask,
                    torch.ones([self.batch_size, 1]).type_as(self.attention_mask),
                ],
                1,
            ).to(device)

        self.past = []

        if isinstance(output[1], tuple):  # past in torch output is tuple
            self.past = list(output[1])
        else:
            for i in range(self.n_layer):
                past_i = (
                    torch.from_numpy(output[i + 1])
                    if isinstance(output[i + 1], numpy.ndarray)
                    else output[i + 1].clone().detach()
                )
                self.past.append(past_i.to(device))

    def diff(self, baseline):
        """
        Compare inputs and logits output.
        """

        print("start diff...")
        if self.logits is not None:
            max_io_diff = (self.logits - baseline.logits).abs().max()
            if max_io_diff > 1e-4:
                print(f"Max logits difference is too large: {max_io_diff}")

        if not torch.all(self.input_ids == baseline.input_ids):
            print("Input_ids is different", self.input_ids, baseline.input_ids)

        if self.has_position_ids:
            if not torch.all(self.position_ids == baseline.position_ids):
                print(
                    "position_ids is different",
                    self.position_ids,
                    baseline.position_ids,
                )

        if self.has_attention_mask:
            if not torch.all(self.attention_mask == baseline.attention_mask):
                print(
                    "attention_mask is different",
                    self.attention_mask,
                    baseline.attention_mask,
                )

        assert len(self.past) == len(baseline.past)

        for i, past_i in enumerate(self.past):
            assert past_i.shape == baseline.past[i].shape
            if past_i.nelement() > 0:
                max_past_diff = (past_i - baseline.past[i]).abs().max()
                if max_past_diff > 1e-4:
                    print(f"max_past_diff[{i}]={max_past_diff}")

    @staticmethod
    def predict_next_token(logits, top_k=1, required_order=False):
        """
        Get top k topkens based on logits.
        """

        # logits has shape (batch_size, seq_len, vocab_size)
        # last token logits has shape (batch_size, vocab_size)
        lastTokenLogits = logits[:, -1]  # noqa: N806
        if top_k == 1:
            generatedTokens = torch.argmax(lastTokenLogits, 1, True)  # noqa: N806
            return generatedTokens
        else:
            topk = torch.argsort(lastTokenLogits, -1, descending=True)[:, :top_k]
            if not required_order:
                sorted_topk, _ = topk.sort()
                return sorted_topk
            return topk

    @staticmethod
    def diff_present(onnx_output, onnx_io_output, n_layer):
        """
        Compare the present outputs of two outputs from ONNX Runtime.
        """
        present_diff_max = []
        for i in range(n_layer):
            onnx_present_i = (
                torch.from_numpy(onnx_output[i + 1])
                if isinstance(onnx_output[i + 1], numpy.ndarray)
                else onnx_output[i + 1]
            )
            onnx_io_present_i = (
                torch.from_numpy(onnx_io_output[i + 1])
                if isinstance(onnx_io_output[i + 1], numpy.ndarray)
                else onnx_io_output[i + 1]
            )
            max_diff = (onnx_present_i - onnx_io_present_i).abs().max()
            present_diff_max.append(max_diff)
        print(f"present_diff_max={present_diff_max}")

    @staticmethod
    def is_quantized_onnx_model(onnx_model_path):
        """
        Returns True if the ONNX model is quantized.
        """
        from onnx import load

        model = load(onnx_model_path)
        from onnxruntime.quantization.quantize import __producer__ as quantize_producer

        return model.producer_name == quantize_producer

    @staticmethod
    def test_generation(
        session,
        model,
        device,
        test_inputs,
        precision=Precision.FLOAT32,
        model_class="Gpt2LMHeadModel",
        top_k=20,
        top_k_no_order=True,
        max_steps=24,
        max_inputs=0,
        verbose=False,
        save_test_data=0,
        save_test_data_dir=".",
    ):
        """
        Test Generation using greedy beam search (without sampling) to compare PyTorch and ONNX model.
        It will print top 1 and top k errors on the given test inputs.
        """
        print(
            f"start test generation: (top_k={top_k} top_k_no_order={top_k_no_order} max_steps={max_steps} test_inputs={len(test_inputs)} max_inputs={max_inputs})"
        )
        n_layer = model.config.n_layer
        n_head = model.config.n_head
        n_embd = model.config.n_embd
        eos_token_id = model.config.eos_token_id
        test_data_saved = 0

        is_float16 = precision == Precision.FLOAT16
        if is_float16:
            assert "float16" in session.get_outputs()[0].type

        # We will still use fp32 torch model as baseline when onnx model if fp16
        model.eval().to(device)

        # Allocate initial buffers for IO Binding of ONNX Runtimne. The buffer size will automatically increase later.
        init_output_shapes = Gpt2Helper.get_output_shapes(
            batch_size=4,
            past_sequence_length=128,
            sequence_length=32,
            config=model.config,
            model_class=model_class,
        )
        output_buffers = Gpt2Helper.get_output_buffers(init_output_shapes, device, is_float16=is_float16)

        baseline_name = "Torch"
        treatment_name = "Quantized Onnx" if precision == Precision.INT8 else "Onnx"
        torch_metric = Gpt2Metric(baseline_name, baseline_name, top_k)
        onnx_metric = Gpt2Metric(treatment_name, baseline_name, top_k)
        onnx_io_metric = Gpt2Metric(treatment_name + " with IO Binding", baseline_name, top_k)

        for i, inputs in enumerate(test_inputs):
            if max_inputs > 0 and i == max_inputs:
                break
            if i % 10 == 0:
                print(f"{i}")
            input_ids = inputs["input_ids"]
            position_ids = inputs.get("position_ids", None)
            attention_mask = inputs.get("attention_mask", None)

            onnx_runner = Gpt2Tester(
                input_ids,
                position_ids,
                attention_mask,
                n_head,
                n_embd,
                n_layer,
                device,
                is_float16,
                top_k,
                not top_k_no_order,
            )
            onnx_io_runner = Gpt2Tester(
                input_ids,
                position_ids,
                attention_mask,
                n_head,
                n_embd,
                n_layer,
                device,
                is_float16,
                top_k,
                not top_k_no_order,
            )
            torch_runner = Gpt2Tester(
                input_ids,
                position_ids,
                attention_mask,
                n_head,
                n_embd,
                n_layer,
                device,
                False,
                top_k,
                not top_k_no_order,
            )  # Torch model baseline is fp32

            batch_size = torch_runner.batch_size
            onnx_metric.start_batch(batch_size)
            onnx_io_metric.start_batch(batch_size)

            with torch.no_grad():
                done = torch.zeros(batch_size, dtype=torch.bool)
                for step in range(max_steps):
                    seq_len = list(onnx_runner.input_ids.size())[1]
                    past_seq_len = list(onnx_runner.past[0].size())[3]

                    start_time = timeit.default_timer()
                    pytorch_output = Gpt2Helper.pytorch_inference(model, torch_runner.get_inputs())
                    torch_metric.add_latency(past_seq_len, timeit.default_timer() - start_time)
                    torch_runner.update(pytorch_output, step, device)

                    onnx_output, avg_latency_ms = Gpt2Helper.onnxruntime_inference(
                        session, onnx_runner.get_inputs(), total_runs=1
                    )
                    onnx_metric.add_latency(past_seq_len, avg_latency_ms / 1000.0)
                    onnx_runner.update(onnx_output, step, device)

                    output_shapes = Gpt2Helper.get_output_shapes(
                        batch_size,
                        past_seq_len,
                        seq_len,
                        model.config,
                        model_class=model_class,
                    )
                    Gpt2Helper.auto_increase_buffer_size(output_buffers, output_shapes)

                    (
                        onnx_io_output,
                        avg_latency_ms,
                    ) = Gpt2Helper.onnxruntime_inference_with_binded_io(
                        session,
                        onnx_io_runner.get_inputs(),
                        output_buffers,
                        output_shapes,
                        total_runs=1,
                        return_numpy=False,
                        include_copy_output_latency=True,
                    )
                    onnx_io_metric.add_latency(past_seq_len, avg_latency_ms / 1000.0)

                    if test_data_saved < save_test_data:
                        onnx_io_runner.save_test_data(session, onnx_io_output, save_test_data_dir, test_data_saved)
                        test_data_saved += 1

                    onnx_io_runner.update(onnx_io_output, step, device)

                    if verbose:
                        onnx_runner.diff(onnx_io_runner)
                        Gpt2Tester.diff_present(onnx_output, onnx_io_output, n_layer)

                        print("Top 1 tokens:")
                        print("\tTorch", torch_runner.top_1_tokens)
                        print("\tONNX", onnx_runner.top_1_tokens)
                        print("\tONNX with IO binding", onnx_io_runner.top_1_tokens)

                    onnx_metric.eval_batch(torch_runner, onnx_runner, past_seq_len, verbose=verbose)
                    onnx_io_metric.eval_batch(torch_runner, onnx_io_runner, past_seq_len, verbose=verbose)

                    done = done | (torch_runner.top_1_tokens == eos_token_id).any()
                    if torch.all(done):
                        break

            onnx_metric.end_batch()
            onnx_io_metric.end_batch()

        torch_metric.print()
        onnx_metric.print()
        onnx_io_metric.print()



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import tempfile
import unittest
from importlib.util import find_spec
from pathlib import Path

import numpy as np
import onnx
from onnx import TensorProto, helper
from op_test_utils import TestDataFeeds, check_model_correctness, check_op_type_count, check_qtype_by_node_type

from onnxruntime.quantization import quant_utils


class TestOpMatMul4Bits(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._tmp_model_dir = tempfile.TemporaryDirectory(prefix="test_matmul4bits.")

    @classmethod
    def tearDownClass(cls):
        cls._tmp_model_dir.cleanup()

    def fill_int4_data(self, shape: int | tuple[int, ...], symmetric: bool) -> np.ndarray:
        line = np.zeros(shape)
        line = line.reshape(-1)

        if symmetric:
            v = -2.0
            for i in range(line.shape[0]):
                if v == 0 or v == -3 or v == 3:
                    v += 1
                line[i] = v
                v += 1
                if v >= 8:
                    v = -8
        else:
            v = 0.0
            for i in range(line.shape[0]):
                line[i] = v
                v += 1
                if v >= 16:
                    v = 0

        return line.reshape(shape)

    def input_feeds(
        self,
        n: int,
        name2shape: dict[str, int | tuple[int, ...]],
        low: int = -1,
        high: int = 2,
        dtype: type = np.float32,
    ) -> TestDataFeeds:
        input_data_list = []
        for _i in range(n):
            inputs = {}
            for name, shape in name2shape.items():
                inputs.update({name: np.random.randint(low, high, shape).astype(dtype)})
            input_data_list.extend([inputs])
        dr = TestDataFeeds(input_data_list)
        return dr

    def construct_model_matmul(self, output_model_path: str, symmetric: bool) -> None:
        #      (input)
        #         |
        #       MatMul
        #         |
        #      (output)
        input_name = "input"
        output_name = "output"
        initializers = []

        def make_matmul(
            input_name, weight_shape: int | tuple[int, ...], weight_name: str, output_name: str, node_name: str
        ):
            weight_data = self.fill_int4_data(weight_shape, symmetric).astype(np.float32)
            initializers.append(onnx.numpy_helper.from_array(weight_data, name=weight_name))
            return onnx.helper.make_node(
                "MatMul",
                [input_name, weight_name],
                [output_name],
                node_name,
            )

        in_features = 52
        out_features = 288
        # make MatMul node
        matmul_node = make_matmul(
            input_name,
            [in_features, out_features],
            "linear1.weight",
            output_name,
            "MatMul_0",
        )

        # make graph
        input_tensor = helper.make_tensor_value_info(input_name, TensorProto.FLOAT, [-1, in_features])
        output_tensor = helper.make_tensor_value_info(output_name, TensorProto.FLOAT, [-1, out_features])
        graph_name = "matmul_4bits_test"
        graph = helper.make_graph(
            [matmul_node],
            graph_name,
            [input_tensor],
            [output_tensor],
            initializer=initializers,
        )
        # blocked quantization requires DQ op set >= 21
        model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 21)])
        model.ir_version = 10  # use stable onnx ir version

        onnx.save(model, output_model_path)

    def construct_model_gather(
        self,
        output_model_path: str,
        symmetric: bool,
        tdata: TensorProto.DataType,
        tind: TensorProto.DataType,
        vocab_size: int = 545,
        embedding_len: int = 228,
    ) -> None:
        #      (input)
        #         |
        #       Gather
        #         |
        #      (output)
        indices_name = "input"
        output_name = "output"
        initializers = []

        def make_gather(
            indices_name, data_shape: int | tuple[int, ...], data_name: str, output_name: str, node_name: str
        ):
            weight_data = self.fill_int4_data(data_shape, symmetric).astype(
                np.float32 if tdata == TensorProto.FLOAT else np.float16
            )
            initializers.append(onnx.numpy_helper.from_array(weight_data, name=data_name))
            kwargs = {"axis": 0}
            return onnx.helper.make_node(
                "Gather",
                [data_name, indices_name],
                [output_name],
                node_name,
                **kwargs,
            )

        gather_node = make_gather(
            indices_name,
            (vocab_size, embedding_len),
            "linear1.weight",
            output_name,
            "Gather_0",
        )

        # make graph
        input_tensor = helper.make_tensor_value_info(indices_name, tind, [-1, 1000])
        output_tensor = helper.make_tensor_value_info(output_name, tdata, [-1, 1000, embedding_len])
        graph_name = "gather_4bits_test"
        graph = helper.make_graph(
            [gather_node],
            graph_name,
            [input_tensor],
            [output_tensor],
            initializer=initializers,
        )
        # QDQ and gather requires op set >= 21. The tool should automatically update the opset.
        model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 19)])
        model.ir_version = 9  # use stable onnx ir version

        onnx.save(model, output_model_path)

    def quant_test(
        self,
        model_fp32_path: str,
        data_reader: TestDataFeeds,
        block_size: int,
        is_symmetric: bool,
        quant_format: quant_utils.QuantFormat = quant_utils.QuantFormat.QOperator,
        op_types_to_quantize: tuple[str, ...] = ("MatMul",),
        quant_axes: tuple[tuple[str, int], ...] = (("MatMul", 0), ("Gather", 1)),
        rtol: float = 0.01,
        atol: float = 0.05,
    ):
        use_qdq = quant_format == quant_utils.QuantFormat.QDQ
        name_prefix = "QDQ" if use_qdq else "QOperator"
        model_int4_path = str(
            Path(self._tmp_model_dir.name).joinpath(f"{name_prefix}_{block_size}_{is_symmetric}.onnx").absolute()
        )

        # Quantize fp32 model to int4 model
        from onnxruntime.quantization import matmul_4bits_quantizer

        model = quant_utils.load_model_with_shape_infer(Path(model_fp32_path))
        quant_config = matmul_4bits_quantizer.DefaultWeightOnlyQuantConfig(
            block_size=block_size,
            is_symmetric=is_symmetric,
            quant_format=quant_format,
            op_types_to_quantize=op_types_to_quantize,
            quant_axes=quant_axes,
        )
        quant = matmul_4bits_quantizer.MatMul4BitsQuantizer(model, algo_config=quant_config)
        quant.process()
        quant.model.save_model_to_file(model_int4_path, False)

        if "Gather" in op_types_to_quantize:
            quant_nodes = {"GatherBlockQuantized": 1}
        else:
            quant_nodes = {"DequantizeLinear": 1, "MatMul": 1} if use_qdq else {"MatMulNBits": 1}
        check_op_type_count(self, model_int4_path, **quant_nodes)

        if use_qdq:
            dq_qtype = TensorProto.INT4 if is_symmetric else TensorProto.UINT4
            dqnode_io_qtypes = (
                {
                    "DequantizeLinear": [
                        ["i", 0, dq_qtype],
                    ]
                }
                if is_symmetric
                else {
                    "DequantizeLinear": [
                        ["i", 0, dq_qtype],
                        ["i", 2, dq_qtype],
                    ]
                }
            )
            check_qtype_by_node_type(self, model_int4_path, dqnode_io_qtypes)
            for op in quant.model.opset_import():
                if op.domain in [None, "", "ai.onnx"] and op.version < 21:
                    self.fail(f"In QDQ format {op.domain} opset should be >= 21")

        data_reader.rewind()

        try:
            check_model_correctness(self, model_fp32_path, model_int4_path, data_reader.get_next(), rtol, atol)
        except Exception as exception:
            if "4b quantization not yet supported on this hardware platform!" in exception.args[0]:
                # Currently we don't have int4 quantization support on all platforms, has to tolerate this exception
                pass
            else:
                raise exception

    def quant_test_with_algo(
        self,
        algorithm: str,
        model_fp32_path: str,
        data_reader: TestDataFeeds,
        block_size: int,
        is_symmetric: bool,
    ):
        model_int4_path = str(
            Path(self._tmp_model_dir.name).joinpath(f"MatMulNBits_{block_size}_{is_symmetric}.onnx").absolute()
        )

        # Quantize fp32 model to int4 model
        from onnxruntime.quantization import matmul_4bits_quantizer

        algo_config = None
        if algorithm == "RTN":
            # test RTN algorithm
            algo_config = matmul_4bits_quantizer.RTNWeightOnlyQuantConfig()
        elif algorithm == "GPTQ":
            # test GPTQ algorithm
            algo_config = matmul_4bits_quantizer.GPTQWeightOnlyQuantConfig(calibration_data_reader=data_reader)
        elif algorithm == "HQQ":
            # test HQQ algorithm
            algo_config = matmul_4bits_quantizer.HQQWeightOnlyQuantConfig(block_size=block_size)

        model = quant_utils.load_model_with_shape_infer(Path(model_fp32_path))
        quant = matmul_4bits_quantizer.MatMul4BitsQuantizer(model, block_size, is_symmetric, algo_config=algo_config)
        quant.process()
        quant.model.save_model_to_file(model_int4_path, False)

        quant_nodes = {"MatMulNBits": 1}
        check_op_type_count(self, model_int4_path, **quant_nodes)

        data_reader.rewind()

        try:
            check_model_correctness(self, model_fp32_path, model_int4_path, data_reader.get_next())
        except Exception as exception:
            if "4b quantization not yet supported on this hardware platform!" in exception.args[0]:
                # Currently we don't have int4 quantization support on all platforms, has to tolerate this exception
                pass
            else:
                raise exception

    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_4bits"
    )
    def test_quantize_matmul_int4_symmetric(self):
        np.random.seed(13)

        model_fp32_path = str(Path(self._tmp_model_dir.name).joinpath("matmul_fp32_symmetric.onnx").absolute())
        self.construct_model_matmul(model_fp32_path, symmetric=True)
        data_reader = self.input_feeds(1, {"input": (100, 52)})
        self.quant_test(model_fp32_path, data_reader, 32, True)

    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_4bits"
    )
    def test_quantize_matmul_int4_offsets(self):
        model_fp32_path = str(Path(self._tmp_model_dir.name).joinpath("matmul_fp32_offset.onnx").absolute())
        self.construct_model_matmul(model_fp32_path, symmetric=False)
        data_reader = self.input_feeds(1, {"input": (100, 52)})
        self.quant_test(model_fp32_path, data_reader, 32, False)

    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_4bits"
    )
    def test_quantize_gather_int4_symmetric(self):
        np.random.seed(13)

        model_fp32_path = str(Path(self._tmp_model_dir.name).joinpath("gather_fp32_symmetric.onnx").absolute())
        self.construct_model_gather(model_fp32_path, True, TensorProto.FLOAT, TensorProto.INT32)
        data_reader = self.input_feeds(1, {"input": (100, 1000)}, -545, 535, np.int32)
        # cover rounding error
        self.quant_test(model_fp32_path, data_reader, 32, True, op_types_to_quantize=("Gather",), rtol=0.2, atol=0.5)

    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_4bits"
    )
    def test_quantize_gather_int4_offsets(self):
        model_fp32_path = str(Path(self._tmp_model_dir.name).joinpath("gather_fp32_offset.onnx").absolute())
        self.construct_model_gather(model_fp32_path, False, TensorProto.FLOAT16, TensorProto.INT64)
        data_reader = self.input_feeds(1, {"input": (100, 1000)}, -545, 535, np.int64)
        # cover rounding error
        self.quant_test(model_fp32_path, data_reader, 32, False, op_types_to_quantize=("Gather",), rtol=0.2, atol=0.5)

    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_4bits"
    )
    def test_quantize_matmul_int4_symmetric_qdq(self):
        np.random.seed(13)

        model_fp32_path = str(Path(self._tmp_model_dir.name).joinpath("matmul_fp32_symmetric.onnx").absolute())
        self.construct_model_matmul(model_fp32_path, symmetric=True)
        data_reader = self.input_feeds(1, {"input": (100, 52)})
        self.quant_test(model_fp32_path, data_reader, 32, True, quant_utils.QuantFormat.QDQ)

    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_4bits"
    )
    def test_quantize_matmul_int4_offsets_qdq(self):
        model_fp32_path = str(Path(self._tmp_model_dir.name).joinpath("matmul_fp32_offset.onnx").absolute())
        self.construct_model_matmul(model_fp32_path, symmetric=False)
        data_reader = self.input_feeds(1, {"input": (100, 52)})
        self.quant_test(model_fp32_path, data_reader, 32, False, quant_utils.QuantFormat.QDQ)

    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_4bits"
    )
    def test_quantize_matmul_int4_using_rtn_algo(self):
        if not find_spec("neural_compressor"):
            self.skipTest("skip test_smooth_quant since neural_compressor is not installed")
        model_fp32_path = str(Path(self._tmp_model_dir.name).joinpath("matmul_fp32_offset.onnx").absolute())
        self.construct_model_matmul(model_fp32_path, symmetric=False)
        data_reader = self.input_feeds(1, {"input": (100, 52)})
        self.quant_test_with_algo("RTN", model_fp32_path, data_reader, 32, False)

    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_4bits"
    )
    def test_quantize_matmul_int4_using_gptq_algo(self):
        if not find_spec("neural_compressor"):
            self.skipTest("skip test_smooth_quant since neural_compressor is not installed")
        model_fp32_path = str(Path(self._tmp_model_dir.name).joinpath("matmul_fp32_offset.onnx").absolute())
        self.construct_model_matmul(model_fp32_path, symmetric=False)
        data_reader = self.input_feeds(1, {"input": (100, 52)})
        self.quant_test_with_algo("GPTQ", model_fp32_path, data_reader, 32, False)

    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_4bits"
    )
    def test_quantize_matmul_int4_using_hqq_algo(self):
        if not find_spec("torch"):
            self.skipTest("skip test_hqq_quant since torch is not installed")
        model_fp32_path = str(Path(self._tmp_model_dir.name).joinpath("matmul_fp32_offset.onnx").absolute())
        self.construct_model_matmul(model_fp32_path, symmetric=False)
        data_reader = self.input_feeds(1, {"input": (100, 52)})
        self.quant_test_with_algo("HQQ", model_fp32_path, data_reader, 32, False)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import tempfile
import unittest
from importlib.util import find_spec
from pathlib import Path

import numpy as np
import onnx
from onnx import TensorProto, helper
from op_test_utils import (
    StridedDataReader,
    check_model_correctness,
    generate_random_initializer,
    input_feeds_neg_one_zero_one,
    input_feeds_neg_one_zero_one_list,
)

import onnxruntime as ort
from onnxruntime.quantization import QuantType, StaticQuantConfig, quantize, quantize_static


def construct_test_model(test_model_path, channel_size):
    """ Create an ONNX model:
    ```
        (input)
          / \
         /   \
      Conv1   \
        |      \
      Relu   Conv3
        |      |
      Conv2    |
        \\      /
           Add
            |
          output
    ```
    """

    input_value_info = helper.make_tensor_value_info("input", TensorProto.FLOAT, [1, channel_size, 1, 3])
    output_value_info = helper.make_tensor_value_info("output", TensorProto.FLOAT, [1, channel_size, 1, 3])

    initializer = []
    initializer.append(generate_random_initializer("W1", [channel_size, channel_size, 1, 1], np.float32))
    initializer.append(generate_random_initializer("B1", [channel_size], np.float32))
    initializer.append(generate_random_initializer("W2", [channel_size, channel_size, 1, 1], np.float32))
    initializer.append(generate_random_initializer("B2", [channel_size], np.float32))
    initializer.append(generate_random_initializer("W3", [channel_size, channel_size, 1, 1], np.float32))
    initializer.append(generate_random_initializer("B3", [channel_size], np.float32))

    nodes = []
    nodes.append(helper.make_node("Conv", ["input", "W1", "B1"], ["conv1_output"], name="conv1"))
    nodes.append(helper.make_node("Relu", ["conv1_output"], ["relu_output"], name="relu"))
    nodes.append(helper.make_node("Conv", ["relu_output", "W2", "B2"], ["conv2_output"], name="conv2"))
    nodes.append(helper.make_node("Conv", ["input", "W3", "B3"], ["conv3_output"], name="conv3"))
    nodes.append(helper.make_node("Add", ["conv2_output", "conv3_output"], ["output"], name="add"))

    graph = helper.make_graph(nodes, "test_graph_4", [input_value_info], [output_value_info], initializer=initializer)
    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 13)])
    onnx.save(model, test_model_path)


class TestStaticQuantization(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # TODO: there will be a refactor to handle all those temporary directories.
        cls._tmp_model_dir = tempfile.TemporaryDirectory(prefix="ort.quant.save.as.external")
        cls._channel_size = 16
        cls._model_fp32_path = str(Path(cls._tmp_model_dir.name) / "fp32.onnx")
        construct_test_model(cls._model_fp32_path, cls._channel_size)

    @classmethod
    def tearDownClass(cls):
        cls._tmp_model_dir.cleanup()

    def test_save_as_external(self):
        data_reader = input_feeds_neg_one_zero_one(10, {"input": [1, self._channel_size, 1, 3]})
        for use_external_data_format in [True, False]:
            quant_model_path = str(Path(self._tmp_model_dir.name) / f"quant.{use_external_data_format}.onnx")
            quantize_static(
                self._model_fp32_path,
                quant_model_path,
                data_reader,
                activation_type=QuantType.QUInt8,
                weight_type=QuantType.QUInt8,
                use_external_data_format=use_external_data_format,
            )

            data_reader.rewind()
            check_model_correctness(self, self._model_fp32_path, quant_model_path, data_reader.get_next())
            data_reader.rewind()

    def run_inference(self, model_path, input_data):
        session = ort.InferenceSession(model_path)
        input_name = session.get_inputs()[0].name
        output_name = session.get_outputs()[0].name
        result = session.run([output_name], {input_name: input_data})
        return result

    def test_stride_effect_on_data_collection(self):
        # Define the stride and test quantize_static with different stride values
        stride = 5
        input_shapes = [1, self._channel_size, 1, 3]
        data_list = input_feeds_neg_one_zero_one_list(10, {"input": [1, self._channel_size, 1, 3]}, 123)
        input_nodes = ["input"]
        in_dtypes = [np.float32]  # Example dtype, adjust as needed

        # strided calibration
        quant_model_path_1 = str(Path(self._tmp_model_dir.name) / "quant.strided.onnx")
        data_reader_1 = StridedDataReader(
            data_list, input_nodes, input_shapes, no_tensor_num=0, in_dtypes=in_dtypes, stride=stride
        )
        quant_config_1 = StaticQuantConfig(data_reader_1, extra_options={"CalibStridedMinMax": stride})
        quantize(str(self._model_fp32_path), str(quant_model_path_1), quant_config_1)

        # non-strided calibration
        quant_model_path_2 = str(Path(self._tmp_model_dir.name) / "quant.non.strided.onnx")
        data_reader_2 = input_feeds_neg_one_zero_one(10, {"input": [1, self._channel_size, 1, 3]}, 123)
        quant_config_2 = StaticQuantConfig(data_reader_2)
        quantize(str(self._model_fp32_path), str(quant_model_path_2), quant_config_2)

        # Inference with both models and assert output closeness
        np.random.seed(123)
        input_data = np.random.choice([-1, 0, 1], size=[1, self._channel_size, 1, 3]).astype(np.float32)

        result_1 = self.run_inference(quant_model_path_1, input_data)
        result_2 = self.run_inference(quant_model_path_2, input_data)

        # Assert that the outputs are close
        np.testing.assert_allclose(
            result_1,
            result_2,
            rtol=0.01,
            atol=0.01,
            err_msg="Outputs from strided and non-strided models are not close enough.",
        )

    def test_static_quant_config(self):
        data_reader = input_feeds_neg_one_zero_one(10, {"input": [1, self._channel_size, 1, 3]})
        quant_config = StaticQuantConfig(data_reader)
        quant_model_path = str(Path(self._tmp_model_dir.name) / "quant.config.onnx")
        quantize(self._model_fp32_path, quant_model_path, quant_config)

        data_reader.rewind()
        check_model_correctness(self, self._model_fp32_path, quant_model_path, data_reader.get_next())
        data_reader.rewind()

    @unittest.skip(
        "Skip failed test in Python Packaging Test Pipeline."
        "During importing neural_compressor, pycocotools throws ValueError: numpy.ndarray size changed"
    )
    def test_smooth_quant(self):
        if not find_spec("neural_compressor"):
            self.skipTest("skip test_smooth_quant since neural_compressor is not installed")
        data_reader = input_feeds_neg_one_zero_one(10, {"input": [1, self._channel_size, 1, 3]})
        quant_config = StaticQuantConfig(data_reader, extra_options={"SmoothQuant": True})
        quant_model_path = str(Path(self._tmp_model_dir.name) / "quant.config.onnx")
        quantize(self._model_fp32_path, quant_model_path, quant_config)

        data_reader.rewind()
        check_model_correctness(self, self._model_fp32_path, quant_model_path, data_reader.get_next())
        data_reader.rewind()

        model = onnx.load(quant_model_path)
        self.assertIn("Mul", [i.op_type for i in model.graph.node])


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from pathlib import Path

import numpy as np
import onnx
import torch

import onnxruntime
from onnxruntime.training.experimental import export_gradient_graph


class NeuralNet(torch.nn.Module):
    r"""
    Simple example model.
    """

    def __init__(self, input_size: int, embedding_size: int, hidden_size: int, num_classes: int):
        super().__init__()

        self.frozen_layer = torch.nn.Linear(input_size, embedding_size, bias=False)
        # Freeze a layer (mainly to test that gradients don't get output for it).
        self.frozen_layer.requires_grad_(False)

        self.fc1 = torch.nn.Linear(embedding_size, hidden_size)
        self.relu = torch.nn.ReLU()
        self.fc2 = torch.nn.Linear(hidden_size, num_classes)

    def forward(self, x):
        out = self.frozen_layer(x)
        out = self.fc1(out)
        out = self.relu(out)
        out = self.fc2(out)
        return out


def to_numpy(tensor):
    return tensor.detach().cpu().numpy() if tensor.requires_grad else tensor.cpu().numpy()


def binary_cross_entropy_loss(inp, target):
    loss = -torch.sum(target * torch.log2(inp[:, 0]) + (1 - target) * torch.log2(inp[:, 1]))
    return loss


class GradientGraphBuilderTest(unittest.TestCase):
    def test_save(self):
        # We need a custom loss function to load the graph in an InferenceSession in ONNX Runtime Web.
        # You can still make the gradient graph with torch.nn.CrossEntropyLoss() and this test will pass.
        loss_fn = binary_cross_entropy_loss
        input_size = 10
        model = NeuralNet(input_size=input_size, embedding_size=20, hidden_size=5, num_classes=2)
        directory_path = Path(os.path.dirname(__file__)).resolve()

        gradient_graph_path = directory_path / "gradient_graph_model.onnx"

        batch_size = 1
        example_input = torch.randn(batch_size, input_size, requires_grad=True)
        example_labels = torch.tensor([1])

        export_gradient_graph(model, loss_fn, example_input, example_labels, gradient_graph_path)

        onnx_model = onnx.load(str(gradient_graph_path))
        onnx.checker.check_model(onnx_model)

        # Expected inputs: input, labels, models parameters.
        self.assertEqual(1 + 1 + sum(1 for _ in model.parameters()), len(onnx_model.graph.input))

        # Expected outputs: prediction, loss, and parameters with gradients.
        self.assertEqual(
            1 + 1 + sum(1 if p.requires_grad else 0 for p in model.parameters()), len(onnx_model.graph.output)
        )

        torch_out = model(example_input)

        try:
            ort_session = onnxruntime.InferenceSession(str(gradient_graph_path))
        except ValueError:
            # Sometimes it is required to pass the available providers.
            from onnxruntime.capi import _pybind_state as C

            available_providers = C.get_available_providers()
            ort_session = onnxruntime.InferenceSession(str(gradient_graph_path), providers=available_providers)

        ort_inputs = {
            onnx_model.graph.input[0].name: to_numpy(example_input),
            onnx_model.graph.input[1].name: to_numpy(example_labels),
        }

        for name, param in model.named_parameters():
            ort_inputs[name] = to_numpy(param.data)

        ort_outs = ort_session.run(None, ort_inputs)
        onnx_output_names = [node.name for node in onnx_model.graph.output]
        onnx_name_to_output = dict(zip(onnx_output_names, ort_outs, strict=False))

        ort_output = onnx_name_to_output["output"]
        np.testing.assert_allclose(to_numpy(torch_out), ort_output, rtol=1e-03, atol=1e-05)

        torch_loss = loss_fn(torch_out, example_labels)
        ort_loss = onnx_name_to_output["loss"]
        np.testing.assert_allclose(to_numpy(torch_loss), ort_loss, rtol=1e-03, atol=1e-05)

        # Make sure the gradients have the right shape.
        model_param_names = tuple(name for name, param in model.named_parameters() if param.requires_grad)
        self.assertEqual(4, len(model_param_names))

        for name, param in model.named_parameters():
            if param.requires_grad:
                grad = onnx_name_to_output[name + "_grad"]
                self.assertEqual(param.size(), grad.shape)


if __name__ == "__main__":
    unittest.main()



import os
import sys  # noqa: F401

import onnx
from onnx import OperatorSetIdProto, TensorProto, helper  # noqa: F401

# Edge that needs to be cut for the split.
# If the edge is feeding into more than one nodes, and not all the nodes belong to the same cut,
# specify those consuming nodes that need to be cut


class CutEdge:
    def __init__(self, edgeId, consumingNodes=None):
        self.edgeId = edgeId
        self.consumingNodes = consumingNodes


def add_expand_type(model, name, type):
    expand_edge = model.graph.value_info.add()
    expand_edge.name = name
    expand_edge.type.CopyFrom(type)


# Add wait/record/send/recv nodes and split the graph into disconnected subgraphs


def split_graph(model, split_edge_groups):
    ms_domain = "com.microsoft"

    new_send_nodes = []
    new_recv_nodes = []

    for cut_index in range(len(split_edge_groups)):
        edgeIds = split_edge_groups[cut_index]  # noqa: N806

        # split the graph based on edgeIds
        upstream_nodes = []
        upstream_nodes_output_index = []
        output_shapes = []
        element_types = []
        for id in edgeIds:
            for node in model.graph.node:
                if len(node.output) >= 1:
                    for i, j in enumerate(node.output):
                        if j == id:
                            upstream_nodes.append(node)
                            upstream_nodes_output_index.append(i)
                            # assuming all tensors are of type float
                            element_types.append(1)
            for info in model.graph.value_info:
                if info.name == id:
                    output_shapes.append(info.type)

        send_input_signal_name = "send_input_signal" + str(cut_index)
        send_signal = model.graph.input.add()
        send_signal.CopyFrom(helper.make_tensor_value_info(send_input_signal_name, onnx.TensorProto.BOOL, None))
        send_signal = helper.make_tensor(send_input_signal_name, TensorProto.BOOL, (), (True,))
        model.graph.initializer.extend([send_signal])

        recv_input_signal_name = "recv_input_signal" + str(cut_index)
        recv_signal = model.graph.input.add()
        recv_signal.CopyFrom(helper.make_tensor_value_info(recv_input_signal_name, onnx.TensorProto.BOOL, None))
        recv_signal = helper.make_tensor(recv_input_signal_name, TensorProto.BOOL, (), (True,))
        model.graph.initializer.extend([recv_signal])

        send_dst_rank_name = "send_dst_rank" + str(cut_index)
        send_dst_rank = model.graph.input.add()
        send_dst_rank.CopyFrom(helper.make_tensor_value_info(send_dst_rank_name, onnx.TensorProto.INT64, None))
        send_dst_rank = helper.make_tensor(send_dst_rank_name, TensorProto.INT64, (), (cut_index + 1,))
        model.graph.initializer.extend([send_dst_rank])

        recv_src_rank_name = "recv_src_rank" + str(cut_index)
        recv_src_rank = model.graph.input.add()
        recv_src_rank.CopyFrom(helper.make_tensor_value_info(recv_src_rank_name, onnx.TensorProto.INT64, None))
        recv_src_rank = helper.make_tensor(recv_src_rank_name, TensorProto.INT64, (), (cut_index,))
        model.graph.initializer.extend([recv_src_rank])

        # output signal from send after cut
        send_output_signal = model.graph.output.add()
        send_output_signal.CopyFrom(
            helper.make_tensor_value_info("send_output_signal" + str(cut_index), onnx.TensorProto.BOOL, None)
        )

        # output signal from receive after cut
        receive_output_signal = model.graph.output.add()
        receive_output_signal.CopyFrom(
            helper.make_tensor_value_info("receive_output_signal" + str(cut_index), onnx.TensorProto.BOOL, None)
        )

        new_send = model.graph.node.add()
        new_send.CopyFrom(
            helper.make_node(
                "Send",
                inputs=[send_input_signal_name, send_dst_rank_name],
                outputs=["send_output_signal" + str(cut_index)],
                tag=0,
                domain=ms_domain,
                element_types=element_types,
                name="send",
            )
        )

        new_receive = model.graph.node.add()
        new_receive.CopyFrom(
            helper.make_node(
                "Recv",
                inputs=[recv_input_signal_name, recv_src_rank_name],
                outputs=["receive_output_signal" + str(cut_index)],
                tag=0,
                domain=ms_domain,
                element_types=element_types,
                name="receive",
            )
        )

        for i in range(len(upstream_nodes)):
            n = upstream_nodes[i]
            idx = upstream_nodes_output_index[i]
            output_type = output_shapes[i]
            output_edge_name = n.output[idx]

            output_nodes = find_all_output_nodes_by_edge(model, output_edge_name)

            # deal with shape inference for newly added edge
            new_send_input_name = output_edge_name + "_send" + str(cut_index)
            add_expand_type(model, new_send_input_name, output_type)

            new_receive_output_name = output_edge_name + "_recv" + str(cut_index)
            add_expand_type(model, new_receive_output_name, output_type)

            # the order of data flow is: node-output -> record -> send -> recv -> wait -> node-input

            new_send.input.extend([output_edge_name])
            new_receive.output.extend([new_receive_output_name])

            for output_node in output_nodes:
                for i in range(len(output_node.input)):  # noqa: PLW2901
                    for edgeId in edgeIds:  # noqa: N806
                        if output_node.input[i] == edgeId:
                            output_node.input[i] = new_receive_output_name

        new_send_nodes.append(new_send)
        new_recv_nodes.append(new_receive)

    model = onnx.shape_inference.infer_shapes(model)

    return new_send_nodes, new_recv_nodes


def find_all_input_nodes(model, node):
    nodes = []
    inputs = []

    if node:
        for inputId in node.input:  # noqa: N806
            nodes.extend([n for n in model.graph.node if inputId in n.output])
            inputs.extend([n for n in model.graph.input if inputId in n.name])
    return nodes, inputs


def find_all_output_nodes(model, node):
    nodes = []
    outputs = []
    if node:
        for outputId in node.output:  # noqa: N806
            nodes.extend([n for n in model.graph.node if outputId in n.input])
            outputs.extend([n for n in model.graph.output if outputId in n.name])
    return nodes, outputs


def find_all_output_nodes_by_edge(model, arg):
    result = [n for n in model.graph.node if arg in n.input]
    return result


# Insert identity nodes to separate same output edge which feeds into different sub-graph.


def add_identity(model, cuttingEdge, newEdgeIdName):
    output_nodes = None
    edgeId = cuttingEdge.edgeId  # noqa: N806
    for node in model.graph.node:
        if len(node.output) >= 1:
            for output in node.output:
                if output == edgeId:
                    output_nodes = find_all_output_nodes_by_edge(model, output)
                    break

    assert output_nodes, "no output node"

    new_identity = model.graph.node.add()
    new_identity.op_type = "Identity"

    new_identity.input.extend([edgeId])
    new_identity.output.extend([newEdgeIdName])

    for i in range(len(output_nodes)):
        for output in output_nodes[i].output:
            if output in cuttingEdge.consumingNodes:
                for j in range(len(output_nodes[i].input)):
                    if output_nodes[i].input[j] == edgeId:
                        output_nodes[i].input[j] = newEdgeIdName

    return new_identity


def insert_identity(model, all_cut_inputs):
    count = 0
    updated_edges = {}
    new_added_identity = []
    split_edge_groups = []
    need_shape_inference = False
    # Sweep the cut edge to see if there are edges feeding into nodes from two sub-graphs. If so,
    # insert identity node after those edges with a new ID to distinguish the rest.
    for cut_input in all_cut_inputs:
        split_edges = []
        for i in cut_input:
            if i.consumingNodes:
                # if this edge has previously been modified, update its edgeId before inserting new identity
                if i.edgeId in updated_edges:
                    i.edgeId = updated_edges[i.edgeId]

                new_edge_name = "identity_output_" + str(count)
                new_added_identity.append(add_identity(model, i, new_edge_name))
                count += 1
                split_edges.append(new_edge_name)
                updated_edges[i.edgeId] = new_edge_name
                need_shape_inference = True
            else:
                split_edges.append(i.edgeId)
        split_edge_groups.append(split_edges)
    return split_edge_groups, new_added_identity, need_shape_inference


# after the graph is split, remove the added identity node because identity op is not registered in gradient builder.


def remove_identity(model, new_added_identity):
    for node in new_added_identity:
        assert node.op_type == "Identity"
        output_nodes = [n for n in model.graph.node if node.output[0] in n.input]
        for output_node in output_nodes:
            for i in range(len(output_node.input)):
                if output_node.input[i] == node.output[0]:
                    output_node.input[i] = node.input[0]


def find_all_connected_nodes(model, node):
    nodes0, inputs = find_all_input_nodes(model, node)
    nodes1, outputs = find_all_output_nodes(model, node)

    connected_nodes = nodes0 + nodes1
    return connected_nodes, inputs, outputs


def get_index(node_list, node):
    found = [i for i, n in enumerate(node_list) if n == node]
    return found[0] if found else None


def get_identity_index_for_deleting(node_list, node):
    for i, n in enumerate(node_list):
        # The node's input name has been changed during send/recv insertion,
        # but it is sufficient to just compare the type and outputs.
        if n.op_type == "Identity" and n.output == node.output:
            return i
    return None


# traverse the graph, group connected nodes and generate subgraph


def generate_subgraph(model, start_nodes, identity_node_list):
    subgraphs = []

    main_graph = onnx.ModelProto()
    main_graph.CopyFrom(model)

    # remove added identity node before copy to subgraph
    identity_node_index = []
    for n in identity_node_list:
        identity_node_index.append(get_identity_index_for_deleting(main_graph.graph.node, n))
    identity_node_index.sort(reverse=True)

    for i in reversed(range(len(main_graph.graph.node))):
        try:
            if i in identity_node_index:
                del main_graph.graph.node[i]
        except Exception:
            print("error deleting identity node", i)

    all_visited_nodes = []
    model_count = len(start_nodes)
    for start in reversed(start_nodes):
        stack0 = [start]

        visited0 = []
        tranversed_node = 0
        inputs0 = []
        outputs0 = []
        while stack0:
            node = stack0.pop()
            if node not in visited0:
                tranversed_node += 1
                visited0.append(node)
                all_visited_nodes.append(node)
                connected_nodes, inputs, outputs = find_all_connected_nodes(main_graph, node)

                stack0 = stack0 + connected_nodes
                inputs0 = inputs0 + inputs
                outputs0 = outputs0 + outputs

        subgraph = onnx.ModelProto()
        subgraph.CopyFrom(main_graph)

        # gather visited nodes
        visited_nodes = []
        for n in visited0:
            visited_nodes.append(get_index(main_graph.graph.node, n))
        visited_nodes.sort(reverse=True)

        # gather visited inputs
        visited_inputs = []
        for n in inputs0:
            visited_inputs.append(get_index(main_graph.graph.input, n))
        visited_inputs.sort(reverse=True)

        # gather visited outputs
        visited_outputs = []
        for n in outputs0:
            visited_outputs.append(get_index(main_graph.graph.output, n))
        visited_outputs.sort(reverse=True)

        for i in reversed(range(len(main_graph.graph.node))):
            try:
                if i not in visited_nodes:
                    del subgraph.graph.node[i]
                else:
                    del main_graph.graph.node[i]
            except Exception:
                print("error deleting node", i)

        for i in reversed(range(len(main_graph.graph.input))):
            try:
                if i not in visited_inputs:
                    del subgraph.graph.input[i]
                else:
                    del main_graph.graph.input[i]
            except Exception:
                print("error deleting inputs", i)

        for i in reversed(range(len(main_graph.graph.output))):
            try:
                if i not in visited_outputs:
                    del subgraph.graph.output[i]
                else:
                    del main_graph.graph.output[i]
            except Exception:
                print("error deleting outputs ", i)

        print("model", str(model_count), " length ", len(subgraph.graph.node))
        subgraphs.append(subgraph)
        model_count -= 1

    print("model", str(model_count), " length ", len(main_graph.graph.node))
    subgraphs.append(main_graph)

    # as the subgraphs were added in reverse order (the last split is added first), reverse the order back before return
    subgraphs.reverse()
    return subgraphs


def main():
    # temporary hard coded the cutting edge structure
    # TODO: move this info to a file (json?) and load the data from there.
    input_model_name = "bert-tiny-uncased_L_3_H_128_A_2_V_30528_S_512_Dp_0.1.onnx"
    stage_count = 3

    cut0_input = {CutEdge("186"), CutEdge("71", {"273", "395"})}
    cut1_input = {CutEdge("308"), CutEdge("71", {"395"})}
    all_cut_inputs = [cut0_input, cut1_input]

    model = onnx.load(input_model_name)
    if len(model.graph.value_info) == 0:
        model = onnx.shape_inference.infer_shapes(model)

    print("original model length ", len(model.graph.node))

    output_model_names = [os.path.splitext(input_model_name)[0] + "_" + str(i) + ".onnx" for i in range(stage_count)]

    split_edge_groups, new_identity, need_shape_inference = insert_identity(model, all_cut_inputs)

    # new edge is being added, need to re-inference shape
    if need_shape_inference:
        model = onnx.shape_inference.infer_shapes(model)

    # after all need-to-be-cut edges identified, split the graph
    new_sends, new_receives = split_graph(model, split_edge_groups)
    remove_identity(model, new_identity)
    sub_graphs = generate_subgraph(model, new_receives, new_identity)

    for i in range(stage_count):
        sub_graphs[i] = onnx.shape_inference.infer_shapes(sub_graphs[i])
        onnx.save(sub_graphs[i], output_model_names[i])
        print("save to file: ", output_model_names[i])


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import os
from enum import IntFlag
from logging import Logger

import torch

from . import _logger, _utils
from ._fallback_exceptions import (
    ORTModuleDeviceException,
    ORTModuleFallbackException,
    ORTModuleInitException,
    ORTModuleIOError,
    ORTModuleONNXModelException,
    ORTModuleTorchModelException,
    wrap_exception,  # noqa: F401
)


class _FallbackPolicy(IntFlag):
    """Policy to trigger fallback from ONNX Runtime engine to PyTorch

    Each policy can be combined with the others (using |) in order to aggregate them
    """

    FALLBACK_DISABLE = 1
    FALLBACK_FORCE_TORCH_FORWARD = 2
    FALLBACK_FORCE_TORCH_BACKWARD = 4
    FALLBACK_UNSUPPORTED_DEVICE = 8
    FALLBACK_UNSUPPORTED_DATA = 16
    FALLBACK_UNSUPPORTED_TORCH_MODEL = 32
    FALLBACK_UNSUPPORTED_ONNX_MODEL = 64
    FALLBACK_BAD_INITIALIZATION = 128

    def is_set(self, policy):
        """Check whether `policy` is set on the `_FallbackPolicy` instance

        FALLBACK_DISABLE implies the check will always fail and return False
        """

        return not self.is_disabled() and policy in self

    def is_disabled(self):
        """Check whether `_FallbackPolicy.FALLBACK_DEVICE` is set on the `_FallbackPolicy` instance"""

        return _FallbackPolicy.FALLBACK_DISABLE in self


class _FallbackManager:
    """Manages fallbacks based on incoming exceptions and specified policies

    The basic algorithm is based on a dictionary whose keys are the supported fallback policies
    and and values are a set of Exception that must be detected.

    When an exception that matches one of the enabled policies are detected,
    a fallback will be pending to execute by ORTModule frontend. If `retry` is False,
    ORTModule will fallback to PyTorch on the following steps. Otherwise, ORTModule [re]try
    to run the model using ORT backend in every step


    On the other hand, when the exception doesn't match any enabled policy, the exception will
    be raised to the user, terminating execution
    """

    def __init__(self, pytorch_module: torch.nn.Module, policy: _FallbackPolicy, retry: bool, logger: Logger):
        self._original_module = pytorch_module

        # Read policy from environment variable for testing purposes
        policy = os.getenv("ORTMODULE_FALLBACK_POLICY", policy)
        if isinstance(policy, str):
            policy = _FallbackPolicy[policy]

        # Read retry from environment variable for testing purposes
        retry = os.getenv("ORTMODULE_FALLBACK_RETRY", str(retry)).lower() in ["true", "1", "yes"]

        self._policy_exception_map = {
            _FallbackPolicy.FALLBACK_FORCE_TORCH_FORWARD.value: {
                ORTModuleFallbackException,
                ORTModuleDeviceException,
                ORTModuleIOError,
                ORTModuleTorchModelException,
                ORTModuleONNXModelException,
            },
            _FallbackPolicy.FALLBACK_FORCE_TORCH_BACKWARD.value: {
                ORTModuleFallbackException,
                ORTModuleDeviceException,
                ORTModuleIOError,
                ORTModuleTorchModelException,
                ORTModuleONNXModelException,
            },
            _FallbackPolicy.FALLBACK_UNSUPPORTED_DEVICE.value: {ORTModuleDeviceException},
            _FallbackPolicy.FALLBACK_UNSUPPORTED_DATA.value: {ORTModuleIOError},
            _FallbackPolicy.FALLBACK_UNSUPPORTED_TORCH_MODEL.value: {ORTModuleTorchModelException},
            _FallbackPolicy.FALLBACK_UNSUPPORTED_ONNX_MODEL.value: {ORTModuleONNXModelException},
            _FallbackPolicy.FALLBACK_BAD_INITIALIZATION.value: {ORTModuleInitException, ORTModuleTorchModelException},
        }
        self.policy = policy
        self.retry = retry
        self._exception = None
        self._raised_fallback_exception = False
        self._logger = logger

    def handle_exception(
        self, exception: Exception, log_level: _logger.LogLevel, override_policy: _FallbackPolicy | None = None
    ) -> None:
        """Process incoming `exception` based on the selected `policy`

        If the incoming `exception` is handled by the specified policy, `_FallbackManager`
        saves the exception so that ORTModule can track the pending fallback
        and trigger it during model execution.
        Otherwise, the incoming exception is immediately raised.

        Args:
            exception (`ORTModuleFallbackException`): Exception that must be handled
            override_policy (`_FallbackPolicy`, optional): Policy to be checked for the incoming `exception`.
                if None is specified, all (except _FallbackPolicy.FALLBACK_DISABLE) are implicitly checked

        Raises:
            `exception`: Original exception is raised when there is no matching policy for it
        """

        def _set_exception(policy: _FallbackPolicy, exception: Exception, log_level: _logger.LogLevel):
            if (
                policy is not _FallbackPolicy.FALLBACK_DISABLE
                and self.policy.is_set(policy)
                and (
                    policy.value in self._policy_exception_map
                    and type(exception) in self._policy_exception_map[policy.value]
                )
            ):
                self._logger.info(f"Fallback for policy {policy.name} is pending.")

                # ORTModuleInitException exceptions do not call `fallback()` through `GraphExecutionManager`,
                # Instead, it fallbacks to PyTorch implicitly through `ORTModule._torch_module = TorchModulePytorch(module)`
                if policy == _FallbackPolicy.FALLBACK_BAD_INITIALIZATION:
                    self._logger.warning(
                        f"Fallback to PyTorch due to exception {type(exception)} was triggered. "
                        "Report this issue with a minimal repro at https://www.github.com/microsoft/onnxruntime. "
                        f"See details below:\n\n{_utils.get_exception_as_string(exception)}"
                    )

                self._exception = exception

        if override_policy is None:
            for policy in _FallbackPolicy:
                _set_exception(policy, exception, log_level)
        else:
            _set_exception(override_policy, exception, log_level)

        if self._exception is None:
            # No fallback, raise failure to user
            raise exception

    def is_pending(self) -> bool:
        """Returns True when a fallback is pending

        ORTModule must execute fallback to PyTorch engine when a pending fallback is detected
        """

        return self._exception is not None

    def fallback(self, log_level: _logger.LogLevel, *inputs, **kwargs):
        """Executes user PyTorch `model` using the provided inputs and return the result"""

        assert self.is_pending(), "`fallback` can only be called when there is a pending fallback"

        if not self._raised_fallback_exception:
            exception_type = type(self._exception)
            exception_string = _utils.get_exception_as_string(self._exception)

            # This warning will not be raised again if retry is not enabled
            self._logger.warning(
                f"Fallback to PyTorch due to exception {exception_type} was triggered. "
                "Report this issue with a minimal repro at https://www.github.com/microsoft/onnxruntime. "
                f"See details below:\n\n{exception_string}"
            )

            self._raised_fallback_exception = True

        # Pending fallbacks are reset to enforce retries
        if self.retry:
            self._raised_fallback_exception = False
            self._exception = None
        return self._original_module(*inputs, **kwargs)



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# For live logging, use the command: pytest -o log_cli=true --log-cli-level=DEBUG

import os
import unittest

import pytest
from test_optimizer import _get_test_model_path


class TestBertProfiler(unittest.TestCase):
    def setUp(self):
        from onnxruntime import get_available_providers

        self.test_cuda = "CUDAExecutionProvider" in get_available_providers()

    def run_profile(self, arguments: str):
        from onnxruntime.transformers.profiler import parse_arguments, run

        args = parse_arguments(arguments.split())
        results = run(args)
        self.assertTrue(len(results) > 1)

    @pytest.mark.slow
    def test_profiler_gpu(self):
        input_model_path = _get_test_model_path("bert_keras_squad")
        if self.test_cuda:
            self.run_profile(f"--model {input_model_path} --batch_size 1 --sequence_length 7 --use_gpu")

    @pytest.mark.slow
    def test_profiler_cpu(self):
        input_model_path = _get_test_model_path("bert_keras_squad")
        self.run_profile(f"--model {input_model_path} --batch_size 1 --sequence_length 7 --dummy_inputs default")


if __name__ == "__main__":
    import sys

    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

    unittest.main()



from pathlib import Path

import onnx
from onnx import TensorProto, helper

# This model contains a MatMul where:
# - A has shape [M, K] and `M` is a dynamic dimension.
# - B is an initializer with shape [K, N].
#   - This is important for the CoreML EP which only handles the case where B is an initializer.

# M is dynamic
M = "M"
K = 2
N = 4

graph = helper.make_graph(
    [  # nodes
        helper.make_node("MatMul", ["A", "B"], ["Y"], "MatMul"),
    ],
    "MatMulWithDynamicInputShape",  # name
    [  # inputs
        helper.make_tensor_value_info("A", TensorProto.FLOAT, [M, K]),
    ],
    [  # outputs
        helper.make_tensor_value_info("Y", TensorProto.FLOAT, [M, N]),
    ],
    [  # initializers
        helper.make_tensor("B", TensorProto.FLOAT, [K, N], [float(i) for i in range(K * N)]),
    ],
)

opset_imports = [helper.make_operatorsetid("", 19)]
model = helper.make_model(graph, opset_imports=opset_imports)
onnx.save(model, str(Path(__file__).parent / "matmul_with_dynamic_input_shape.onnx"))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# _load_config_from_json.py

import json
import logging
import os
from functools import reduce
from types import SimpleNamespace

from onnxruntime.capi import _pybind_state as C
from onnxruntime.training.ortmodule._fallback import _FallbackPolicy
from onnxruntime.training.ortmodule.options import DebugOptions, LogLevel, _SaveOnnxOptions, _SkipCheck

from . import JSON_PATH_ENVIRONMENT_KEY

log = logging.getLogger(__name__)


def _load_data_from_json(path):
    """Loads data from the json file path provided."""

    data = None
    with open(path) as f:
        data = json.load(f, object_hook=lambda d: SimpleNamespace(**d))

    if data is None:
        raise RuntimeError(f"No data found in provided json file {path}.")

    return data


def _load_propagate_cast_ops(ortmodule_config_accessor, data):
    """Loads PropagateCastOps from json file onto ORTModule."""

    assert hasattr(data, _load_propagate_cast_ops.loading_key)
    log.info(f"Found keyword {_load_propagate_cast_ops.loading_key} in json. Loading attributes from file.")

    def _update_strategy():
        ortmodule_config_accessor._runtime_options.propagate_cast_ops_strategy = C.PropagateCastOpsStrategy.__members__[
            data.PropagateCastOps.Strategy
        ]

    def _update_level():
        ortmodule_config_accessor._runtime_options.propagate_cast_ops_level = data.PropagateCastOps.Level

    def _update_allow():
        ortmodule_config_accessor._runtime_options.propagate_cast_ops_allow = data.PropagateCastOps.Allow

    key_to_function_mapping = {"Strategy": _update_strategy, "Level": _update_level, "Allow": _update_allow}

    for key in data.PropagateCastOps.__dict__:
        key_to_function_mapping[key]()


def _load_use_external_gpu_allocator(ortmodule_config_accessor, data):
    """Loads UseExternalGPUAllocator from json file onto ORTModule."""

    assert hasattr(data, _load_use_external_gpu_allocator.loading_key)
    log.info(f"Found keyword {_load_use_external_gpu_allocator.loading_key} in json. Loading attributes from file.")

    assert isinstance(data.UseExternalGPUAllocator, bool), (
        f"{_load_use_external_gpu_allocator.loading_key} must be a boolean"
    )
    ortmodule_config_accessor._runtime_options.use_external_gpu_allocator = data.UseExternalGPUAllocator


def _load_enable_custom_autograd_function(ortmodule_config_accessor, data):
    """Loads EnableCustomAutogradFunction from json file onto ORTModule."""

    assert hasattr(data, _load_enable_custom_autograd_function.loading_key)
    log.info(
        f"Found keyword {_load_enable_custom_autograd_function.loading_key} in json. Loading attributes from file."
    )

    assert isinstance(data.EnableCustomAutogradFunction, bool), (
        f"{_load_enable_custom_autograd_function.loading_key} must be a boolean"
    )

    from onnxruntime.training.ortmodule._custom_autograd_function import enable_custom_autograd_support

    enable_custom_autograd_support(data.EnableCustomAutogradFunction)
    ortmodule_config_accessor._runtime_options.enable_custom_autograd_function = data.EnableCustomAutogradFunction


def _load_enable_grad_acc_optimization(ortmodule_config_accessor, data):
    """Loads EnableGradAccOptimization from json file onto ORTModule."""

    assert hasattr(data, _load_enable_grad_acc_optimization.loading_key)
    log.info(f"Found keyword {_load_enable_grad_acc_optimization.loading_key} in json. Loading attributes from file.")

    assert isinstance(data.EnableGradAccOptimization, bool), (
        f"{_load_enable_grad_acc_optimization.loading_key} must be a boolean"
    )
    ortmodule_config_accessor._runtime_options.enable_grad_acc_optimization = data.EnableGradAccOptimization


def _load_run_symbolic_shape_infer(ortmodule_config_accessor, data):
    """Loads RunSymbolicShapeInference from json file onto ORTModule."""

    assert hasattr(data, _load_run_symbolic_shape_infer.loading_key)
    log.info(f"Found keyword {_load_run_symbolic_shape_infer.loading_key} in json. Loading attributes from file.")

    assert isinstance(data.RunSymbolicShapeInference, bool), (
        f"{_load_run_symbolic_shape_infer.loading_key} must be a boolean"
    )
    ortmodule_config_accessor._runtime_options.run_symbolic_shape_infer = data.RunSymbolicShapeInference


def _load_use_static_shape(ortmodule_config_accessor, data):
    """Loads UseStaticShape from json file onto ORTModule."""

    assert hasattr(data, _load_use_static_shape.loading_key)
    log.info(f"Found keyword {_load_use_static_shape.loading_key} in json. Loading attributes from file.")

    assert isinstance(data.UseStaticShape, bool), f"{_load_use_static_shape.loading_key} must be a boolean"
    ortmodule_config_accessor._runtime_options.use_static_shape = data.UseStaticShape


def _load_skip_check(ortmodule_config_accessor, data):
    """Loads SkipCheck from json file onto ORTModule."""

    assert hasattr(data, _load_skip_check.loading_key)
    log.info(f"Found keyword {_load_skip_check.loading_key} in json. Loading attributes from file.")

    skip_check = reduce(lambda x, y: x | y, [_SkipCheck[name] for name in data.SkipCheck])
    if skip_check.value > 0:
        ortmodule_config_accessor._runtime_options.skip_check = skip_check


def _load_debug_options(ortmodule_config_accessor, data):
    """Loads DebugOptions from json file onto ORTModule."""

    assert hasattr(data, _load_debug_options.loading_key)
    log.info(f"Found keyword {_load_debug_options.loading_key} in json. Loading attributes from file.")

    log_level = LogLevel.WARNING

    def _update_log_level():
        nonlocal log_level
        log_level = LogLevel[data.DebugOptions.LogLevel]

    save_onnx = False

    def _update_save_onnx():
        nonlocal save_onnx
        save_onnx = data.DebugOptions.SaveONNX

    onnx_prefix = ""

    def _update_onnx_prefix():
        nonlocal onnx_prefix
        onnx_prefix = data.DebugOptions.ONNXPrefix

    def _update_onnx_path():
        os.environ[_SaveOnnxOptions._path_environment_key] = data.DebugOptions.SaveONNXPath

    key_to_function_mapping = {
        "LogLevel": _update_log_level,
        "SaveONNX": _update_save_onnx,
        "ONNXPrefix": _update_onnx_prefix,
        "SaveONNXPath": _update_onnx_path,
    }

    for key in data.DebugOptions.__dict__:
        key_to_function_mapping[key]()

    debug_options = DebugOptions(log_level=log_level, save_onnx=save_onnx, onnx_prefix=onnx_prefix)
    ortmodule_config_accessor._debug_options = debug_options


def _load_use_memory_efficient_gradient(ortmodule_config_accessor, data):
    """Loads UseMemoryEfficientGradient from json file onto ORTModule."""

    assert hasattr(data, _load_use_memory_efficient_gradient.loading_key)
    log.info(f"Found keyword {_load_use_memory_efficient_gradient.loading_key} in json. Loading attributes from file.")

    assert isinstance(data.UseMemoryEfficientGradient, bool), (
        f"{_load_use_memory_efficient_gradient.loading_key} must be a boolean"
    )
    ortmodule_config_accessor._runtime_options.use_memory_efficient_gradient = data.UseMemoryEfficientGradient


def _load_fallback_policy(ortmodule_config_accessor, data):
    """Loads SkipCheck from json file onto ORTModule."""

    assert hasattr(data, _load_fallback_policy.loading_key)
    log.info(f"Found keyword {_load_fallback_policy.loading_key} in json. Loading attributes from file.")

    fallback_policy = reduce(lambda x, y: x | y, [_FallbackPolicy[name] for name in data.FallbackPolicy])
    if fallback_policy.value > 0:
        ortmodule_config_accessor._fallback_manager.policy = fallback_policy


def _load_onnx_opset_version(ortmodule_config_accessor, data):
    """Loads OnnxOpsetVersion from json file onto ORTModule."""

    assert hasattr(data, _load_onnx_opset_version.loading_key)
    log.info(f"Found keyword {_load_onnx_opset_version.loading_key} in json. Loading attributes from file.")

    assert isinstance(data.OnnxOpsetVersion, int), f"{_load_onnx_opset_version.loading_key} must be an int"
    ortmodule_config_accessor._runtime_options.onnx_opset_version = data.OnnxOpsetVersion


def _define_load_function_keys():
    """Define static key variables for each loading function"""

    _load_propagate_cast_ops.loading_key = "PropagateCastOps"
    _load_use_external_gpu_allocator.loading_key = "UseExternalGPUAllocator"
    _load_enable_custom_autograd_function.loading_key = "EnableCustomAutogradFunction"
    _load_enable_grad_acc_optimization.loading_key = "EnableGradAccOptimization"
    _load_run_symbolic_shape_infer.loading_key = "RunSymbolicShapeInference"
    _load_use_static_shape.loading_key = "UseStaticShape"
    _load_skip_check.loading_key = "SkipCheck"
    _load_debug_options.loading_key = "DebugOptions"
    _load_use_memory_efficient_gradient.loading_key = "UseMemoryEfficientGradient"
    _load_fallback_policy.loading_key = "FallbackPolicy"
    _load_onnx_opset_version.loading_key = "OnnxOpsetVersion"


def load_from_json(ortmodule, path=None):
    """Loads config from json file at given path.

    Here is the schema that the json file must adhere to:
    {
        "PropagateCastOps":
        {
            "Strategy": "FLOOD_FILL", # str representing strategy ("NONE", "FLOOD_FILL", or "INSERT_AND_REDUCE")
            "Level": 3, # propagate cast ops level as an int
            "Allow": ["ABC", "DEF"] # propagate cast ops allow as list of strs
        },
        "UseExternalGPUAllocator" : false, # bool flag
        "EnableCustomAutogradFunction": true, # bool flag
        "EnableGradAccOptimization": true, # bool flag
        "UseStaticShape": true, # bool flag
        "RunSymbolicShapeInference": false, # bool flag
        "SkipCheck": # list of strs representing `_SkipCheck`s checks to skip which will be aggregated using |
        [
            "SKIP_CHECK_DEVICE",
            "SKIP_CHECK_BUILD_GRADIENT",
            "SKIP_CHECK_EXECUTION_AGENT"
        ],
        "DebugOptions": # debug options for user facing configuration
        {
            "LogLevel": "VERBOSE",
            "SaveONNX": true,
            "ONNXPrefix": "my_model",
            "SaveONNXPath": "/path/to/onnx/directory"
        },
        "FallbackPolicy": # list of strings representing fallback policies (`_FallbackPolicy`s which can be aggregated using |
        [
            "FALLBACK_DISABLE",
            "FALLBACK_FORCE_TORCH_FORWARD",
            "FALLBACK_UNSUPPORTED_DEVICE",
            "FALLBACK_UNSUPPORTED_DATA",
            "FALLBACK_UNSUPPORTED_TORCH_MODEL",
            "FALLBACK_UNSUPPORTED_ONNX_MODEL",
            "FALLBACK_BAD_INITIALIZATION",
        ],
        "OnnxOpsetVersion": 14 # int defining the opset version to be used during export
    }

    Args:
        ortmodule (:obj:`ORTModule`): ORTModule instance that needs to be configured
        path (:obj:`str`, optional): Path to json file. Alternatively, users can set the
            environment variable ORTMODULE_JSON_CONFIG_PATH to the json config path. In case
            both path and environment variable are set, the environment variable gets precedence.
    """

    path = os.getenv(JSON_PATH_ENVIRONMENT_KEY, path)

    # figure out the json path
    if path is None:
        raise ValueError(
            "Path to json is not provided."
            f"Provide the path through function call or setting the environment variable {JSON_PATH_ENVIRONMENT_KEY}"
        )

    # load the entire json file
    data = _load_data_from_json(path)

    # define the keys for all loading functions
    _define_load_function_keys()

    # define all load functions to iterate over
    load_functions = {
        _load_propagate_cast_ops.loading_key: _load_propagate_cast_ops,
        _load_use_external_gpu_allocator.loading_key: _load_use_external_gpu_allocator,
        _load_enable_custom_autograd_function.loading_key: _load_enable_custom_autograd_function,
        _load_enable_grad_acc_optimization.loading_key: _load_enable_grad_acc_optimization,
        _load_run_symbolic_shape_infer.loading_key: _load_run_symbolic_shape_infer,
        _load_use_static_shape.loading_key: _load_use_static_shape,
        _load_skip_check.loading_key: _load_skip_check,
        _load_debug_options.loading_key: _load_debug_options,
        _load_use_memory_efficient_gradient.loading_key: _load_use_memory_efficient_gradient,
        _load_fallback_policy.loading_key: _load_fallback_policy,
        _load_onnx_opset_version.loading_key: _load_onnx_opset_version,
    }

    for training_mode in [True, False]:
        # update the debug config for both train and eval modes
        ortmodule_config_accessor = ortmodule._torch_module._execution_manager(training_mode)
        # iterate over the json data instead of checking for keys in json to catch key errors
        for key in data.__dict__:
            load_functions[key](ortmodule_config_accessor, data)



import numpy as np
import onnx
from onnx import OperatorSetIdProto, TensorProto, helper, numpy_helper

X = helper.make_tensor_value_info("input", TensorProto.FLOAT, ["batch", "seqlen", 128])
unsqueezed_masked_lm_positions = helper.make_tensor_value_info(
    "unsqueezed_masked_lm_positions",
    TensorProto.INT64,
    ["batch", "dynamic_prediction_count", 1],
)
Y = helper.make_tensor_value_info("output", TensorProto.FLOAT, ["batch", "dynamic_prediction_count", 128])
Y2 = helper.make_tensor_value_info("output2", TensorProto.FLOAT, ["batch", "dynamic_prediction_count", 128])
nodes = []

# case 1
bias_np_val = np.random.uniform(0.0, 1.0, (128)).astype(np.float32).reshape(128)
bias_initializer = numpy_helper.from_array(bias_np_val, "bias")
add1 = helper.make_node("Add", ["input", "bias"], ["add_1"], name="add_1")
nodes.append(add1)

gathernd1 = helper.make_node(
    "GatherND",
    ["add_1", "unsqueezed_masked_lm_positions"],
    ["output"],
    name="gathernd_1",
    batch_dims=1,
)
nodes.append(gathernd1)

# case 2
bias2_np_val = np.random.uniform(0.0, 1.0, (128)).astype(np.float32).reshape(128)
bias2_initializer = numpy_helper.from_array(bias2_np_val, "bias2")
add2 = helper.make_node("Add", ["bias2", "input"], ["add_2"], name="add_2")
nodes.append(add2)

gathernd2 = helper.make_node(
    "GatherND",
    ["add_2", "unsqueezed_masked_lm_positions"],
    ["gathernd_out"],
    name="gathernd_2",
    batch_dims=1,
)
nodes.append(gathernd2)

identity = helper.make_node("Identity", ["gathernd_out"], ["output2"], name="identity")
nodes.append(identity)

graph_def = helper.make_graph(
    nodes,
    "test-model",
    [X, unsqueezed_masked_lm_positions],
    [Y, Y2],
    [bias_initializer, bias2_initializer],
)

opsets = []
onnxdomain = OperatorSetIdProto()
onnxdomain.version = 12
onnxdomain.domain = ""  # The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
opsets.append(onnxdomain)

msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"

opsets.append(msdomain)
kwargs = {}
kwargs["opset_imports"] = opsets

model_def = helper.make_model(graph_def, producer_name="onnx-example", **kwargs)

onnx.save(model_def, "gathernd_add.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import math

import numpy
import onnx
from bert_model_generator import float_tensor, reverse_if
from onnx import TensorProto, helper
from packaging import version


def create_gpt2_attention(hidden_size=64, num_heads=4, max_seq_len=32, switch_add_inputs=False):
    # unsqueeze in opset version 13 has two inputs (axis is moved from attribute to input).
    is_opset_13_or_newer = version.parse(onnx.__version__) >= version.parse("1.8.0")

    # nodes in attention subgraph
    nodes = [
        helper.make_node("Add", ["input_1", "input_2"], ["layernorm_input"], "add_layernorm"),
        helper.make_node(
            "LayerNormalization",
            ["layernorm_input", "layer_norm_weight", "layer_norm_bias"],
            ["layernorm_out"],
            "layernorm",
            epsion=0.000009999999747378752,
        ),
        # fully connection nodes
        helper.make_node(
            "MatMul",
            ["layernorm_out", "matmul_fc_weight"],
            ["matmul_fc_out"],
            "matmul_fc",
        ),
        helper.make_node(
            "Add",
            reverse_if(["matmul_fc_out", "add_fc_weight"], switch_add_inputs),
            ["fc_out"],
            "add_fc",
        ),
        (
            helper.make_node("Split", ["fc_out", "split_q_k_v"], ["q", "k", "v"], "split_qkv", axis=2)
            if is_opset_13_or_newer
            else helper.make_node(
                "Split",
                ["fc_out"],
                ["q", "k", "v"],
                "split_qkv",
                axis=2,
                split=[hidden_size, hidden_size, hidden_size],
            )
        ),
        # q nodes
        helper.make_node("Reshape", ["q", "reshape_x_shape"], ["reshape_q_out"], "reshape_q"),
        helper.make_node(
            "Transpose",
            ["reshape_q_out"],
            ["transpose_q_out"],
            "transpose_q",
            perm=[0, 2, 1, 3],
        ),
        # k nodes
        helper.make_node("Reshape", ["k", "reshape_x_shape"], ["reshape_k_out"], "reshape_k"),
        helper.make_node(
            "Transpose",
            ["reshape_k_out"],
            ["transpose_k_out"],
            "transpose_k",
            perm=[0, 2, 1, 3],
        ),
        # v nodes
        helper.make_node("Reshape", ["v", "reshape_x_shape"], ["reshape_v_out"], "reshape_v"),
        helper.make_node(
            "Transpose",
            ["reshape_v_out"],
            ["transpose_v_out"],
            "transpose_v",
            perm=[0, 2, 1, 3],
        ),
        # past
        (
            helper.make_node("Split", ["past", "split_1_1"], ["split_k", "split_v"], "split_past", axis=0)
            if is_opset_13_or_newer
            else helper.make_node(
                "Split",
                ["past"],
                ["split_k", "split_v"],
                "split_past",
                axis=0,
                split=[1, 1],
            )
        ),
        (
            helper.make_node("Squeeze", ["split_k", "axes_0"], ["past_k"], "squeeze_past_k")
            if is_opset_13_or_newer
            else helper.make_node("Squeeze", ["split_k"], ["past_k"], "squeeze_past_k", axes=[0])
        ),
        helper.make_node(
            "Concat",
            ["past_k", "transpose_k_out"],
            ["concat_k_out"],
            "concat_k",
            axis=-2,
        ),
        helper.make_node(
            "Transpose",
            ["concat_k_out"],
            ["concat_k_transpose_out"],
            "transpose_concat_k",
            perm=[0, 1, 3, 2],
        ),
        (
            helper.make_node("Squeeze", ["split_v", "axes_0"], ["past_v"], "squeeze_past_v")
            if is_opset_13_or_newer
            else helper.make_node("Squeeze", ["split_v"], ["past_v"], "squeeze_past_v", axes=[0])
        ),
        helper.make_node(
            "Concat",
            ["past_v", "transpose_v_out"],
            ["concat_v_out"],
            "concat_v",
            axis=-2,
        ),
        # present
        (
            helper.make_node(
                "Unsqueeze",
                ["concat_k_out", "axes_0"],
                ["concat_k_unsqueeze_out"],
                "concat_k_unsqueeze",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Unsqueeze",
                ["concat_k_out"],
                ["concat_k_unsqueeze_out"],
                "concat_k_unsqueeze",
                axes=[0],
            )
        ),
        (
            helper.make_node(
                "Unsqueeze",
                ["concat_v_out", "axes_0"],
                ["concat_v_unsqueeze_out"],
                "concat_v_unsqueeze",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Unsqueeze",
                ["concat_v_out"],
                ["concat_v_unsqueeze_out"],
                "concat_v_unsqueeze",
                axes=[0],
            )
        ),
        helper.make_node(
            "Concat",
            ["concat_k_unsqueeze_out", "concat_v_unsqueeze_out"],
            ["present"],
            "concat_present",
            axis=0,
        ),
        helper.make_node("Shape", ["transpose_q_out"], ["transpose_q_shape_out"], "transpose_q_shape"),
        helper.make_node(
            "Slice",
            ["transpose_q_shape_out", "starts_n2", "ends_n1", "axes_0"],
            ["transpose_q_shape_slice_out"],
            "transpose_q_shape_slice",
        ),
        (
            helper.make_node(
                "Squeeze",
                ["transpose_q_shape_slice_out", "axes_0"],
                ["transpose_q_shape_slice_squeeze_out"],
                "transpose_q_shape_slice_squeeze",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Squeeze",
                ["transpose_q_shape_slice_out"],
                ["transpose_q_shape_slice_squeeze_out"],
                "transpose_q_shape_slice_squeeze",
                axes=[0],
            )
        ),
        helper.make_node("Shape", ["concat_k_out"], ["concat_k_shape_out"], "concat_k_shape"),
        helper.make_node(
            "Slice",
            ["concat_k_shape_out", "starts_n2", "ends_n1", "axes_0"],
            ["concat_k_shape_slice_out"],
            "concat_k_shape_slice",
        ),
        (
            helper.make_node(
                "Squeeze",
                ["concat_k_shape_slice_out", "axes_0"],
                ["concat_k_shape_slice_squeeze_out"],
                "concat_k_shape_slice_squeeze",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Squeeze",
                ["concat_k_shape_slice_out"],
                ["concat_k_shape_slice_squeeze_out"],
                "concat_k_shape_slice_squeeze",
                axes=[0],
            )
        ),
        helper.make_node(
            "Sub",
            ["concat_k_shape_slice_squeeze_out", "transpose_q_shape_slice_squeeze_out"],
            ["sub_out"],
            "sub",
        ),
        (
            helper.make_node("Unsqueeze", ["sub_out", "axes_0"], ["sub_unsqueeze_out"], "sub_unsqueeze")
            if is_opset_13_or_newer
            else helper.make_node("Unsqueeze", ["sub_out"], ["sub_unsqueeze_out"], "sub_unsqueeze", axes=[0])
        ),
        (
            helper.make_node(
                "Unsqueeze",
                ["concat_k_shape_slice_squeeze_out", "axes_0"],
                ["concat_k_shape_slice_squeeze_unsqueeze_out"],
                "concat_k_shape_slice_squeeze_unsqueeze",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Unsqueeze",
                ["concat_k_shape_slice_squeeze_out"],
                ["concat_k_shape_slice_squeeze_unsqueeze_out"],
                "concat_k_shape_slice_squeeze_unsqueeze",
                axes=[0],
            )
        ),
        helper.make_node(
            "Slice",
            [
                "undir_mask",
                "sub_unsqueeze_out",
                "concat_k_shape_slice_squeeze_unsqueeze_out",
                "axes_2",
                "steps_1",
            ],
            ["undir_mask_slice_out"],
            "undir_mask_slice",
        ),
        helper.make_node(
            "Slice",
            [
                "undir_mask_slice_out",
                "starts_0",
                "concat_k_shape_slice_squeeze_unsqueeze_out",
                "axes_3",
                "steps_1",
            ],
            ["mask_slice_slice_out"],
            "mask_slice_slice",
        ),
        helper.make_node(
            "Cast",
            ["mask_slice_slice_out"],
            ["undir_mask_out"],
            "undir_mask_cast",
            to=9,
        ),
        # mask nodes
        helper.make_node(
            "Reshape",
            ["input_mask", "input_mask_shape"],
            ["input_mask_reshape_out"],
            "input_mask_reshape",
        ),
        (
            helper.make_node(
                "Unsqueeze",
                ["input_mask_reshape_out", "axes_1"],
                ["unsqueeze0_out"],
                "unsqueeze0",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Unsqueeze",
                ["input_mask_reshape_out"],
                ["unsqueeze0_out"],
                "unsqueeze0",
                axes=[1],
            )
        ),
        (
            helper.make_node("Unsqueeze", ["unsqueeze0_out", "axes_2"], ["unsqueeze1_out"], "unsqueeze1")
            if is_opset_13_or_newer
            else helper.make_node("Unsqueeze", ["unsqueeze0_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[2])
        ),
        helper.make_node("Sub", ["sub_weight", "unsqueeze1_out"], ["mask_sub_out"], "sub_mask"),
        helper.make_node("Mul", ["mask_sub_out", "mul_weight"], ["mul_mask_out"], "mul_mask"),
        # qk nodes
        helper.make_node(
            "MatMul",
            ["transpose_q_out", "concat_k_transpose_out"],
            ["qk_out"],
            "matmul_qk",
        ),
        helper.make_node("Div", ["qk_out", "div_weight"], ["qk_norm_out"], "qk_norm"),
        helper.make_node(
            "Where",
            ["undir_mask_out", "qk_norm_out", "where_weight"],
            ["where_out"],
            "where",
        ),
        helper.make_node(
            "Add",
            reverse_if(["where_out", "mul_mask_out"], switch_add_inputs),
            ["add_mask_out"],
            "add_mask",
        ),
        helper.make_node("Softmax", ["add_mask_out"], ["softmax_out"], "softmax", axis=3),
        # qkv nodes
        helper.make_node(
            "MatMul",
            ["softmax_out", "concat_v_out"],
            ["matmul_qkv_1_out"],
            "matmul_qk_v",
        ),
        helper.make_node(
            "Transpose",
            ["matmul_qkv_1_out"],
            ["transpose_qkv_out"],
            "transpose_qkv",
            perm=[0, 2, 1, 3],
        ),
        helper.make_node(
            "Reshape",
            ["transpose_qkv_out", "reshape_weight_qkv"],
            ["reshape_qkv_out"],
            "reshape_qkv",
        ),
        helper.make_node("Shape", ["reshape_qkv_out"], ["qkv_shape"], "shape_qkv"),
        helper.make_node(
            "Slice",
            ["qkv_shape", "starts_n1", "ends_inf", "axes_0"],
            ["qkv_shape_slice_out"],
            "qkv_shape_slice",
        ),
        (
            helper.make_node(
                "Squeeze",
                ["qkv_shape_slice_out", "axes_0"],
                ["qkv_shape_slice_squeeze_out"],
                "qkv_shape_slice_squeeze",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Squeeze",
                ["qkv_shape_slice_out"],
                ["qkv_shape_slice_squeeze_out"],
                "qkv_shape_slice_squeeze",
                axes=[0],
            )
        ),
        (
            helper.make_node(
                "Unsqueeze",
                ["qkv_shape_slice_squeeze_out", "axes_0"],
                ["qkv_shape_slice_squeeze_unsqueeze_out"],
                "qkv_shape_slice_squeeze_unsqueeze",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Unsqueeze",
                ["qkv_shape_slice_squeeze_out"],
                ["qkv_shape_slice_squeeze_unsqueeze_out"],
                "qkv_shape_slice_squeeze_unsqueeze",
                axes=[0],
            )
        ),
        helper.make_node(
            "Concat",
            ["concat_n1", "qkv_shape_slice_squeeze_unsqueeze_out"],
            ["qkv_shape_slice_squeeze_unsqueeze_concat_out"],
            "qkv_shape_slice_squeeze_unsqueeze_concat",
            axis=0,
        ),
        helper.make_node(
            "Reshape",
            ["reshape_qkv_out", "qkv_shape_slice_squeeze_unsqueeze_concat_out"],
            ["qkv_reshape_out"],
            "qkv_reshape",
        ),
        helper.make_node(
            "Gemm",
            ["qkv_reshape_out", "gemm_weight", "gemm_bias"],
            ["gemm_out"],
            "gemm",
            alpha=1.0,
            beta=1.0,
            transA=0,
            transB=0,
        ),
        helper.make_node(
            "Gather",
            ["qkv_shape", "indices_1"],
            ["qkv_shape_1"],
            "shape_qkv_gather_1",
            axis=0,
        ),
        helper.make_node(
            "Gather",
            ["qkv_shape", "indices_0"],
            ["qkv_shape_0"],
            "shape_qkv_gather_0",
            axis=0,
        ),
        (
            helper.make_node(
                "Unsqueeze",
                ["qkv_shape_1", "axes_0"],
                ["qkv_shape_1_unsqueeze_out"],
                "qkv_shape_1_unsqueeze",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Unsqueeze",
                ["qkv_shape_1"],
                ["qkv_shape_1_unsqueeze_out"],
                "qkv_shape_1_unsqueeze",
                axes=[0],
            )
        ),
        (
            helper.make_node(
                "Unsqueeze",
                ["qkv_shape_0", "axes_0"],
                ["qkv_shape_0_unsqueeze_out"],
                "qkv_shape_0_unsqueeze",
            )
            if is_opset_13_or_newer
            else helper.make_node(
                "Unsqueeze",
                ["qkv_shape_0"],
                ["qkv_shape_0_unsqueeze_out"],
                "qkv_shape_0_unsqueeze",
                axes=[0],
            )
        ),
        helper.make_node(
            "Concat",
            ["qkv_shape_0_unsqueeze_out", "qkv_shape_1_unsqueeze_out", "qkv_hidden"],
            ["shape_qkv_concat_out"],
            "shape_qkv_concat",
            axis=0,
        ),
        helper.make_node(
            "Reshape",
            ["gemm_out", "shape_qkv_concat_out"],
            ["gemm_reshape_out"],
            "gemm_reshape",
        ),
        helper.make_node(
            "Add",
            reverse_if(["gemm_reshape_out", "layernorm_input"], switch_add_inputs),
            ["skip_output"],
            "add_skip",
        ),
        helper.make_node(
            "LayerNormalization",
            ["skip_output", "layer_norm_weight", "layer_norm_bias"],
            ["output"],
            "layernorm2",
            epsion=0.000009999999747378752,
        ),
    ]

    head_size = int(hidden_size // num_heads)
    unidir_mask = (
        numpy.tril(numpy.ones((max_seq_len, max_seq_len))).reshape([max_seq_len * max_seq_len]).astype(numpy.uint8)
    )
    initializers = [  # initializers
        float_tensor("layer_norm_weight", [hidden_size]),
        float_tensor("layer_norm_bias", [hidden_size]),
        float_tensor("matmul_fc_weight", [hidden_size, 3 * hidden_size]),
        float_tensor("add_fc_weight", [3 * hidden_size]),
        float_tensor("gemm_weight", [hidden_size, hidden_size]),
        float_tensor("gemm_bias", [hidden_size]),
        helper.make_tensor(
            "undir_mask",
            TensorProto.UINT8,
            [1, 1, max_seq_len, max_seq_len],
            unidir_mask.tolist(),
        ),
        helper.make_tensor("div_weight", TensorProto.FLOAT, [], [math.sqrt(head_size)]),
        helper.make_tensor("sub_weight", TensorProto.FLOAT, [], [1.0]),
        helper.make_tensor("where_weight", TensorProto.FLOAT, [], [-10000.0]),
        helper.make_tensor("mul_weight", TensorProto.FLOAT, [], [-10000]),
        helper.make_tensor("input_mask_shape", TensorProto.INT64, [2], [0, -1]),
        helper.make_tensor("starts_0", TensorProto.INT64, [1], [0]),
        helper.make_tensor("concat_n1", TensorProto.INT64, [1], [-1]),
        helper.make_tensor("starts_n1", TensorProto.INT64, [1], [-1]),
        helper.make_tensor("ends_inf", TensorProto.INT64, [1], [9223372036854775807]),
        helper.make_tensor("starts_n2", TensorProto.INT64, [1], [-2]),
        helper.make_tensor("ends_n1", TensorProto.INT64, [1], [-1]),
        helper.make_tensor("axes_0", TensorProto.INT64, [1], [0]),
        helper.make_tensor("axes_2", TensorProto.INT64, [1], [2]),
        helper.make_tensor("axes_3", TensorProto.INT64, [1], [3]),
        helper.make_tensor("steps_1", TensorProto.INT64, [1], [1]),
        helper.make_tensor("indices_0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices_1", TensorProto.INT64, [], [1]),
        helper.make_tensor("qkv_hidden", TensorProto.INT64, [1], [hidden_size]),
        helper.make_tensor("reshape_x_shape", TensorProto.INT64, [4], [0, 0, num_heads, head_size]),
        helper.make_tensor("reshape_weight_qkv", TensorProto.INT64, [3], [0, 0, hidden_size]),
    ]

    if is_opset_13_or_newer:
        initializers.append(helper.make_tensor("split_1_1", TensorProto.INT64, [2], [1, 1]))
        initializers.append(
            helper.make_tensor(
                "split_q_k_v",
                TensorProto.INT64,
                [3],
                [hidden_size, hidden_size, hidden_size],
            )
        )
        initializers.append(helper.make_tensor("axes_1", TensorProto.INT64, [1], [1]))

    graph = helper.make_graph(
        [node for node in nodes if node],
        "GPT2",  # name
        [  # inputs
            helper.make_tensor_value_info(
                "input_1",
                TensorProto.FLOAT,
                ["batch_size", "sequence_length", hidden_size],
            ),
            helper.make_tensor_value_info(
                "input_2",
                TensorProto.FLOAT,
                ["batch_size", "sequence_length", hidden_size],
            ),
            helper.make_tensor_value_info(
                "input_mask",
                TensorProto.FLOAT,
                ["batch_size", "past_sequence_length + sequence_length"],
            ),
            helper.make_tensor_value_info(
                "past",
                TensorProto.FLOAT,
                [2, "batch_size", num_heads, "past_sequence_length", head_size],
            ),
        ],
        [  # outputs
            helper.make_tensor_value_info(
                "output",
                TensorProto.FLOAT,
                ["batch_size", "sequence_length", hidden_size],
            ),
            helper.make_tensor_value_info(
                "present",
                TensorProto.FLOAT,
                [
                    2,
                    "batch_size",
                    num_heads,
                    "past_sequence_length + sequence_length",
                    head_size,
                ],
            ),
        ],
        initializers,
    )

    # Needed so that we don't see the new LayerNormalization function added in version 17.
    # TODO(https://github.com/microsoft/onnxruntime/issues/11916): Remove once fixed.
    opsetid = helper.make_opsetid("ai.onnx", min(onnx.defs.onnx_opset_version(), 16))
    return helper.make_model(graph, opset_imports=(opsetid,))


def create_gpt2_embedlayer(
    pos_embed=100,
    word_embed=101,
    token_embed=10,
    hidden_size=768,
    attn_hidden_dim=256,
    num_heads=4,
    epsilon=0.1,
    one_attention_node=False,
    has_skip_layer_norm=True,
    output_embedding_sum=False,
):
    # Construct input and output nodes
    inputs = [
        helper.make_tensor_value_info("ids", TensorProto.INT32, ["batch_size", "sequence_length"]),
    ]
    outputs = [
        helper.make_tensor_value_info("output_0", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden_size])
    ]

    if output_embedding_sum:
        outputs.append(
            helper.make_tensor_value_info(
                "embedding_sum", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden_size]
            )
        )

    # Construct graph nodes
    embed_layernorm_nodes = [
        helper.make_node("Gather", ["word_embeddings_weight", "ids"], ["gather_0_out"], "gather_word_embeddings"),
        helper.make_node("Gather", ["pos_embeddings_weight", "ids"], ["gather_1_out"], "gather_position_embeddings"),
        helper.make_node("Add", ["gather_0_out", "gather_1_out"], ["add_0_out"], "add_before_layernorm"),
        helper.make_node("Gather", ["token_embeddings_weight", "ids"], ["gather_2_out"], "gather_token_embeddings"),
    ]

    if has_skip_layer_norm:
        embed_layernorm_nodes.append(
            helper.make_node(
                "SkipLayerNormalization",
                ["add_0_out", "gather_2_out", "layernorm_weight", "layernorm_bias"],
                ["skip_layernorm_out"] if not output_embedding_sum else ["skip_layernorm_out", "", "", "embedding_sum"],
                "skip_layernorm",
                domain="com.microsoft",
                epsilon=epsilon,
            )
        )
    else:
        embed_layernorm_nodes.append(
            helper.make_node("Add", ["add_0_out", "gather_2_out"], ["embedding_sum"], "embedding_sum")
        )

        embed_layernorm_nodes.append(
            helper.make_node(
                "LayerNormalization",
                ["embedding_sum", "layernorm_weight", "layernorm_bias"],
                ["skip_layernorm_out"],
                "layernorm",
                epsilon=epsilon,
            )
        )

    attention_nodes = (
        [
            helper.make_node("MatMul", ["skip_layernorm_out", "q_weight"], ["q_out"], "q_attn"),
            helper.make_node("MatMul", ["skip_layernorm_out", "k_weight"], ["k_out"], "k_attn"),
            helper.make_node("MatMul", ["skip_layernorm_out", "v_weight"], ["v_out"], "v_attn"),
            helper.make_node("Add", ["q_out", "k_out"], ["qk_out"], "qk_attn"),
            helper.make_node("Add", ["qk_out", "v_out"], ["qkv_out"], "qkv_attn"),
        ]
        if not one_attention_node
        else [
            helper.make_node(
                "Attention",
                ["skip_layernorm_out", "qkv_weight", "qkv_bias", ""],
                ["attn_out"],
                "qkv_attn",
                domain="com.microsoft",
                num_heads=num_heads,
            ),
            helper.make_node(
                "MatMul",
                ["attn_out", "fix_hidden_size"],
                ["qkv_out"],
                "matmul_after_attn",
            ),
        ]
    )
    nodes = [
        helper.make_node(
            "SkipLayerNormalization",
            ["skip_layernorm_out", "qkv_out", "layernorm_weight", "layernorm_bias", "dense_bias"],
            ["output_0"],
            "attn_skip_layernorm",
            domain="com.microsoft",
            epsilon=epsilon,
        ),
    ]
    nodes.extend(embed_layernorm_nodes)
    nodes.extend(attention_nodes)

    # Construct data initializers for graph nodes
    embed_layernorm_initializers = [
        helper.make_tensor(
            "word_embeddings_weight",
            TensorProto.FLOAT,
            [word_embed, hidden_size],
            [(i + 1) / (word_embed * hidden_size) for i in range(word_embed * hidden_size)],
        ),
        helper.make_tensor(
            "pos_embeddings_weight",
            TensorProto.FLOAT,
            [pos_embed, hidden_size],
            [(i + 2) / (pos_embed * hidden_size) for i in range(pos_embed * hidden_size)],
        ),
        helper.make_tensor(
            "token_embeddings_weight",
            TensorProto.FLOAT,
            [token_embed, hidden_size],
            [(i + 3) / (token_embed * hidden_size) for i in range(token_embed * hidden_size)],
        ),
        helper.make_tensor(
            "layernorm_weight", TensorProto.FLOAT, [hidden_size], [(i + 4) / hidden_size for i in range(hidden_size)]
        ),
        helper.make_tensor(
            "layernorm_bias", TensorProto.FLOAT, [hidden_size], [(i + 5) / hidden_size for i in range(hidden_size)]
        ),
    ]
    attention_initializers = (
        [
            helper.make_tensor(
                "q_weight",
                TensorProto.FLOAT,
                [hidden_size, hidden_size],
                [(i + 6) / (hidden_size * hidden_size) for i in range(hidden_size * hidden_size)],
            ),
            helper.make_tensor(
                "k_weight",
                TensorProto.FLOAT,
                [hidden_size, hidden_size],
                [(i + 7) / (hidden_size * hidden_size) for i in range(hidden_size * hidden_size)],
            ),
            helper.make_tensor(
                "v_weight",
                TensorProto.FLOAT,
                [hidden_size, hidden_size],
                [(i + 8) / (hidden_size * hidden_size) for i in range(hidden_size * hidden_size)],
            ),
        ]
        if not one_attention_node
        else [
            helper.make_tensor(
                "qkv_weight",
                TensorProto.FLOAT,
                [hidden_size, hidden_size],
                [(i + 9) / (hidden_size * hidden_size) for i in range(hidden_size * hidden_size)],
            ),
            helper.make_tensor(
                "qkv_bias", TensorProto.FLOAT, [hidden_size], [(i + 10) / hidden_size for i in range(hidden_size)]
            ),
            helper.make_tensor(
                "fix_hidden_size",
                TensorProto.FLOAT,
                [attn_hidden_dim, hidden_size],
                [(i + 11) / (attn_hidden_dim * hidden_size) for i in range(attn_hidden_dim * hidden_size)],
            ),
        ]
    )
    initializers = [
        helper.make_tensor(
            "dense_bias", TensorProto.FLOAT, [hidden_size], [(i + 12) / hidden_size for i in range(hidden_size)]
        ),
    ]
    initializers.extend(embed_layernorm_initializers)
    initializers.extend(attention_initializers)

    # Construct graph
    graph = helper.make_graph(nodes, "GPT2_embedlayer_graph", inputs, outputs, initializers)
    opsetid = helper.make_opsetid("ai.onnx", min(onnx.defs.onnx_opset_version(), 16))
    return helper.make_model(graph, opset_imports=(opsetid,))


def create_gpt2_fused_embedlayer(
    pos_embed=100,
    word_embed=101,
    token_embed=10,
    hidden_size=768,
    attn_hidden_dim=256,
    num_heads=4,
    epsilon=0.1,
    one_attention_node=False,
    output_embedding_sum=False,
):
    # Construct input and output nodes
    inputs = [
        helper.make_tensor_value_info("ids", TensorProto.INT32, ["batch_size", "sequence_length"]),
    ]
    outputs = [
        helper.make_tensor_value_info("output_0", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden_size])
    ]
    if output_embedding_sum:
        outputs.append(
            helper.make_tensor_value_info(
                "embedding_sum", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden_size]
            )
        )

    # Construct graph nodes
    embed_layernorm_nodes = [
        helper.make_node(
            "EmbedLayerNormalization",
            [
                "ids",
                "ids",
                "word_embeddings_weight",
                "pos_embeddings_weight",
                "token_embeddings_weight",
                "layernorm_weight",
                "layernorm_bias",
                "",
                "ids",
            ],
            (
                ["EmbedLayerNormalization_0_output", "EmbedLayerNormalization_0_dummy_mask_index", "embedding_sum"]
                if output_embedding_sum
                else ["EmbedLayerNormalization_0_output", "EmbedLayerNormalization_0_dummy_mask_index"]
            ),
            "EmbedLayerNormalization_0",
            domain="com.microsoft",
            epsilon=epsilon,
        ),
    ]
    attention_nodes = (
        [
            helper.make_node("MatMul", ["EmbedLayerNormalization_0_output", "q_weight"], ["q_out"], "q_attn"),
            helper.make_node("MatMul", ["EmbedLayerNormalization_0_output", "k_weight"], ["k_out"], "k_attn"),
            helper.make_node("MatMul", ["EmbedLayerNormalization_0_output", "v_weight"], ["v_out"], "v_attn"),
            helper.make_node("Add", ["q_out", "k_out"], ["qk_out"], "qk_attn"),
            helper.make_node("Add", ["qk_out", "v_out"], ["qkv_out"], "qkv_attn"),
        ]
        if not one_attention_node
        else [
            helper.make_node(
                "Attention",
                ["EmbedLayerNormalization_0_output", "qkv_weight", "qkv_bias", ""],
                ["attn_out"],
                "qkv_attn",
                domain="com.microsoft",
                num_heads=num_heads,
            ),
            helper.make_node(
                "MatMul",
                ["attn_out", "fix_hidden_size"],
                ["qkv_out"],
                "matmul_after_attn",
            ),
        ]
    )
    nodes = [
        helper.make_node(
            "SkipLayerNormalization",
            ["EmbedLayerNormalization_0_output", "qkv_out", "layernorm_weight", "layernorm_bias", "dense_bias"],
            ["output_0"],
            "attn_skip_layernorm",
            domain="com.microsoft",
            epsilon=epsilon,
        ),
    ]
    nodes.extend(embed_layernorm_nodes)
    nodes.extend(attention_nodes)

    # Construct data initializers for graph nodes
    embed_layernorm_initializers = [
        helper.make_tensor(
            "word_embeddings_weight",
            TensorProto.FLOAT,
            [word_embed, hidden_size],
            [(i + 1) / (word_embed * hidden_size) for i in range(word_embed * hidden_size)],
        ),
        helper.make_tensor(
            "pos_embeddings_weight",
            TensorProto.FLOAT,
            [pos_embed, hidden_size],
            [(i + 2) / (pos_embed * hidden_size) for i in range(pos_embed * hidden_size)],
        ),
        helper.make_tensor(
            "token_embeddings_weight",
            TensorProto.FLOAT,
            [token_embed, hidden_size],
            [(i + 3) / (token_embed * hidden_size) for i in range(token_embed * hidden_size)],
        ),
        helper.make_tensor(
            "layernorm_weight", TensorProto.FLOAT, [hidden_size], [(i + 4) / hidden_size for i in range(hidden_size)]
        ),
        helper.make_tensor(
            "layernorm_bias", TensorProto.FLOAT, [hidden_size], [(i + 5) / hidden_size for i in range(hidden_size)]
        ),
    ]
    attention_initializers = (
        [
            helper.make_tensor(
                "q_weight",
                TensorProto.FLOAT,
                [hidden_size, hidden_size],
                [(i + 6) / (hidden_size * hidden_size) for i in range(hidden_size * hidden_size)],
            ),
            helper.make_tensor(
                "k_weight",
                TensorProto.FLOAT,
                [hidden_size, hidden_size],
                [(i + 7) / (hidden_size * hidden_size) for i in range(hidden_size * hidden_size)],
            ),
            helper.make_tensor(
                "v_weight",
                TensorProto.FLOAT,
                [hidden_size, hidden_size],
                [(i + 8) / (hidden_size * hidden_size) for i in range(hidden_size * hidden_size)],
            ),
        ]
        if not one_attention_node
        else [
            helper.make_tensor(
                "qkv_weight",
                TensorProto.FLOAT,
                [hidden_size, hidden_size],
                [(i + 9) / (hidden_size * hidden_size) for i in range(hidden_size * hidden_size)],
            ),
            helper.make_tensor(
                "qkv_bias", TensorProto.FLOAT, [hidden_size], [(i + 10) / hidden_size for i in range(hidden_size)]
            ),
            helper.make_tensor(
                "fix_hidden_size",
                TensorProto.FLOAT,
                [attn_hidden_dim, hidden_size],
                [(i + 11) / (attn_hidden_dim * hidden_size) for i in range(attn_hidden_dim * hidden_size)],
            ),
        ]
    )
    initializers = [
        helper.make_tensor(
            "dense_bias", TensorProto.FLOAT, [hidden_size], [(i + 12) / hidden_size for i in range(hidden_size)]
        ),
    ]
    initializers.extend(embed_layernorm_initializers)
    initializers.extend(attention_initializers)

    # Construct graph
    graph = helper.make_graph(nodes, "GPT2_embedlayer_graph", inputs, outputs, initializers)
    opsetid = helper.make_opsetid("ai.onnx", min(onnx.defs.onnx_opset_version(), 16))
    return helper.make_model(graph, opset_imports=(opsetid,))


if __name__ == "__main__":
    model = create_gpt2_attention()
    onnx.save(model, "gpt2_attention.onnx")

    model = create_gpt2_attention(switch_add_inputs=True)
    onnx.save(model, "gpt2_attention_add.onnx")

    model = create_gpt2_embedlayer()
    onnx.save(model, "gpt2_embedlayer.onnx")

    model = create_gpt2_fused_embedlayer()
    onnx.save(model, "./test_data/models/gpt2_embedlayer_exp.onnx")

    model = create_gpt2_embedlayer(one_attention_node=True)
    onnx.save(model, "gpt2_embedlayer_one_attn.onnx")

    model = create_gpt2_fused_embedlayer(one_attention_node=True)
    onnx.save(model, "./test_data/models/gpt2_embedlayer_one_attn_exp.onnx")

    model = create_gpt2_embedlayer(one_attention_node=True, output_embedding_sum=True)
    onnx.save(model, "gpt2_embedlayer_one_attn_output_sum.onnx")

    model = create_gpt2_fused_embedlayer(one_attention_node=True, output_embedding_sum=True)
    onnx.save(model, "./test_data/models/gpt2_embedlayer_one_attn_output_sum_exp.onnx")
